from pathlib import Path
from typing import Any, Dict, List, Optional

import pandas as pd
from pydantic import validate_call

from energy_consumption_forecasting.exceptions import log_exception
from energy_consumption_forecasting.logger import get_logger

logger = get_logger(name=Path(__file__).name)


@log_exception(logger=logger)
@validate_call(config=dict(arbitrary_types_allowed=True))
def clean_dataframe(
    dataframe: pd.DataFrame,
    check_columns_duplicates: List[Any],
    drop_columns: Optional[List[Any]] = None,
) -> pd.DataFrame:
    """
    This function cleans the provided dataframe by removing missing and duplicate
    data.

    Parameters
    ----------
    dataframe: pd.DataFrame
        The pandas dataframe as a input for transformation.

    check_columns_duplicate: List[Any]
        A list containing column names for checking duplication and dropping the rows.

    drop_columns: List[Any] or None, default=None
        A list containing column names for dropping it from the DataFrame.

    Returns
    -------
    pd.DataFrame
        Returns a cleaned pandas dataframe.
    """
    dataset_df = dataframe.copy()

    # Cleaning the NaN values
    dataset_df.dropna(inplace=True)

    # Dropping unwanted columns
    if drop_columns is not None:
        if len(drop_columns) > 0:
            for col in drop_columns:
                if not col in dataset_df.columns:
                    raise Exception(
                        f'Column "{col}" does not exist in the dataFrame, '
                        "kindly recheck the column names to drop it."
                    )
            dataset_df.drop(columns=drop_columns, inplace=True)
        else:
            raise Exception("Empty list argument is not allowed in drop_columns.")

    # Dropping duplicate values in the dataframe
    dataset_df.drop_duplicates(
        subset=check_columns_duplicates,
        ignore_index=True,
        inplace=True,
    )

    return dataset_df


@log_exception(logger=logger)
@validate_call(config=dict(arbitrary_types_allowed=True))
def rename_features(
    dataframe: pd.DataFrame,
    rename_columns_dict: Dict[Any, Any],
) -> pd.DataFrame:
    """
    This function rename features of the dataframe
    by using the provided new names in dict format.

    Parameters
    ----------
    dataframe: pd.DataFrame
        The pandas dataframe as a input for transformation.

    rename_columns_dict: Dict[Any, Any]
        A dict containing existing column names as keys and new column names as values.

    Returns
    -------
    pd.DataFrame
        Returns a pandas dataframe with new feature names.
    """
    dataset_df = dataframe.copy()

    if bool(rename_columns_dict):
        for key in rename_columns_dict.keys():
            if not key in dataset_df.columns:
                raise Exception(
                    f'Column "{key}" does not exist in the dataFrame, '
                    "kindly recheck the column names to rename it."
                )
        dataset_df.rename(columns=rename_columns_dict, inplace=True)
    else:
        raise Exception("Empty Dict argument is not allowed in rename_columns_dict.")

    return dataset_df


@log_exception(logger=logger)
@validate_call(config=dict(arbitrary_types_allowed=True))
def casting_features(dataframe: pd.DataFrame) -> pd.DataFrame:
    """
    This function transform features data type of the dataframe.

    Parameters
    ----------
    dataframe: pd.DataFrame
        The pandas dataframe as a input for transformation.

    Returns
    -------
    pd.DataFrame
        Returns a pandas dataframe with transformed feature data types.
    """
    dataset_df = dataframe.copy()

    dataset_df["datetime_dk"] = pd.to_datetime(
        dataset_df["datetime_dk"],
        format="%Y-%m-%dT%H:%M:%S",
    )
    dataset_df["branch"] = dataset_df["branch"].astype("string")
    dataset_df["municipality_num"] = dataset_df["municipality_num"].astype("int32")
    dataset_df["consumption_kwh"] = dataset_df["consumption_kwh"].astype("float64")

    return dataset_df


@log_exception(logger=logger)
@validate_call(config=dict(arbitrary_types_allowed=True))
def feature_engineering(dataframe: pd.DataFrame) -> pd.DataFrame:
    """
    This function performs feature engineering on the dataframe.

    Parameters
    ----------
    dataframe: pd.DataFrame
        The pandas dataframe as a input for transformation.

    Returns
    -------
    pd.DataFrame
        Returns a pandas dataframe after performing feature engineering.
    """
    dataset_df = dataframe.copy()

    # Encoding the branch feature categories
    branch_maps = {
        "Offentligt": 1,
        "Erhverv": 2,
        "Privat": 3,
    }
    dataset_df["branch"] = (
        dataset_df["branch"].apply(lambda x: branch_maps.get(x)).astype("int8")
    )

    # Creating a new feature for consumption,
    # converting kilowatt-hour into megawatt-hour

    # dataset_df['consumption_mwh'] = dataset_df['consumption_kwh'] / 1000

    # Creating new columns based on the data present in the feature datetime_dk
    # dataset_df['hour_dk'] = dataset_df.datetime_dk.dt.hour
    # dataset_df['dayofmonth_dk'] = dataset_df.datetime_dk.dt.day
    # dataset_df['month_dk'] = dataset_df.datetime_dk.dt.month
    # dataset_df['year_dk'] = dataset_df.datetime_dk.dt.year
    # dataset_df['dayofweek_dk'] = dataset_df.datetime_dk.dt.dayofweek
    # dataset_df['dayofyear_dk'] = dataset_df.datetime_dk.dt.dayofyear
    # dataset_df['weekofyear'] = dataset_df.datetime_dk.dt.isocalendar().week
    # dataset_df['quarter_dk'] = dataset_df.datetime_dk.dt.quarter

    return dataset_df



from datetime import datetime
from typing import Literal, Optional

import wandb
from pydantic import validate_call

from energy_consumption_forecasting.utils import get_env_var

WANDB_API_KEY = get_env_var(key="WANDB_API_KEY")
WANDB_ENTITY = get_env_var(key="WANDB_ENTITY")
WANDB_PROJECT = get_env_var(key="WANDB_PROJECT")


@validate_call
def init_wandb_run(
    run_name: str,
    add_timestamp_to_run_name: bool = False,
    project_name: str = WANDB_PROJECT,
    entity: str = WANDB_ENTITY,
    group: Optional[str] = None,
    job_type: Optional[str] = None,
    resume: Optional[Literal["auto", "allow", "must", "never"]] = None,
    reinit: bool = False,
    run_id: Optional[str] = None,
) -> Optional[wandb.run]:
    """
    This function creates a wrapper to initiates the WandB experiment
    run with "wandb.init()" function for tracking the experiment.

    This can be used with a context manager like "with" block statement in python.

    Parameters
    ----------
    run_name: str
        A short name for the run, you can identify the run in the UI by this run name.

    add_timestamp_to_run_name: bool, default=False
        A timestamp is added at the end of the run name.

    project_name: str, default=WANDB_PROJECT
        The name of the project where newly created run will be send to, by default
        the project name takes the value from env variable with key "WANDB_PROJECT".

    entity: str, default=WANDB_ENTITY
        An entity is the username or a team name that is generated when an account
        is created in WandB, by default it takes the value from the env variable with
        key "WANDB_ENTITY".

    group: str or None, default=None
        A group name to group different runs into a specific group.

    job_type: str or None, default=None
        A job type to distinguish multiple different runs within the group.

    resume: Literal["auto", "allow", "must", "never"] or None, default=None
        Sets the resuming behavior for the run, only "auto", "allow", "must", "never"
        None are allowed as inputs.

    reinit: bool, default=False
        whether to allow multiple wandb.init() calls in the same process.

    run_id: str or None, default=None
        A unique id for this run , this can be used for resuming. It can not contains
        special characters like "/\#?%:".

    Returns
    -------
    wandb.run or None
        A wrapper around wandb.init() that returns the run object,
        can be used with the "with" block statement and None if
        the "wandb.finish" cleans up the run object.
    """

    if add_timestamp_to_run_name:
        run_name = f"{run_name}_{datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}"

    run = wandb.init(
        name=run_name,
        project=project_name,
        entity=entity,
        group=group,
        job_type=job_type,
        resume=resume,
        reinit=reinit,
        id=run_id,
    )

    return run



from datetime import datetime

import numpy as np
import pandas as pd
import pytest

from energy_consumption_forecasting.feature_pipeline.data_transformation import (
    casting_features,
    clean_dataframe,
    feature_engineering,
    rename_features,
)


# Creating a dummy dataframe for testing the functions
@pytest.fixture
def get_dataframe():
    data = {
        "HourUTC": [
            "2023-06-30T22:00:00",
            "2023-07-01T23:00:00",
            "2023-07-02T00:00:00",
            "2023-06-30T22:00:00",
            "2023-07-04T02:00:00",
        ],
        "HourDK": [
            "2023-07-01T00:00:00",
            "2023-07-02T01:00:00",
            "2023-07-03T02:00:00",
            "2023-07-01T00:00:00",
            "2023-07-05T04:00:00",
        ],
        "MunicipalityNo": [
            250,
            773,
            766,
            250,
            np.NaN,
        ],
        "Branche": [
            "Offentligt",
            "Erhverv",
            "Privat",
            "Offentligt",
            "Erhverv",
        ],
        "ConsumptionkWh": [
            1361.189,
            7812.050,
            6760.285,
            1361.189,
            72003.364,
        ],
    }

    return pd.DataFrame(data)


def test_clean_dataframe(get_dataframe):
    """
    In this test the dataframe is cleaned and returned as pandas dataframe.
    """

    assert get_dataframe.shape == (5, 5)
    result_df = clean_dataframe(dataframe=get_dataframe)
    assert isinstance(result_df, pd.DataFrame)
    assert result_df.shape == (3, 5)

    # Dropping a column
    result_df = clean_dataframe(dataframe=get_dataframe, drop_columns=["HourUTC"])
    assert result_df.shape == (3, 4)

    with pytest.raises(Exception) as exe_info:
        result_df = clean_dataframe(dataframe=get_dataframe, drop_columns=[])
        assert "Empty list argument is not allowed in drop_columns." in str(
            exe_info.value
        )

        result_df = clean_dataframe(dataframe=get_dataframe, drop_columns=[123])
        assert (
            'Column "123" does not exist in the dataFrame, '
            "kindly recheck the column names to drop it." in str(exe_info.value)
        )


def test_rename_features(get_dataframe):
    """
    In this test the dataframe columns are renamed and returned as pandas dataframe.
    """

    rename_dict = {
        "HourDK": "datetime_dk",
        "MunicipalityNo": "municipality_num",
    }
    result_df = rename_features(
        dataframe=get_dataframe, rename_columns_dict=rename_dict
    )
    assert isinstance(result_df, pd.DataFrame)
    assert (
        "HourDK" not in result_df.columns and "MunicipalityNo" not in result_df.columns
    )
    assert (
        "datetime_dk" in result_df.columns and "municipality_num" in result_df.columns
    )

    with pytest.raises(Exception) as exe_info:
        result_df = rename_features(dataframe=get_dataframe, rename_columns_dict={})
        assert "Empty Dict argument is not allowed in rename_columns_dict." in str(
            exe_info.value
        )

        result_df = rename_features(
            dataframe=get_dataframe, rename_columns_dict={"Hour": "datetime"}
        )
        assert (
            'Column "Hour" does not exist in the dataFrame, '
            "kindly recheck the column names to rename it." in str(exe_info.value)
        )


def test_casting_features(get_dataframe):
    """
    In this test the data type of dataframe features are transformed
    and returned as pandas dataframe.
    """

    df = clean_dataframe(dataframe=get_dataframe)
    rename_dict = {
        "HourDK": "datetime_dk",
        "MunicipalityNo": "municipality_num",
        "Branche": "branch",
        "ConsumptionkWh": "consumption_kwh",
    }
    df = rename_features(dataframe=df, rename_columns_dict=rename_dict)
    result_df = casting_features(dataframe=df)

    assert isinstance(result_df, pd.DataFrame)

    assert isinstance(result_df.datetime_dk.dtype, type(np.dtype("datetime64[ns]")))
    assert bool(datetime.strftime(result_df.datetime_dk[2], "%Y-%m-%d %H:%M:%S"))

    assert isinstance(result_df.municipality_num.dtype, type(np.dtype("int32")))
    assert isinstance(result_df.branch.dtype, pd.StringDtype)
    assert isinstance(result_df.consumption_kwh.dtype, type(np.dtype("float64")))


def test_feature_engineering(get_dataframe):
    """
    In this test the dataframe perform feature engineering and returns the dataframe
    """
    df = rename_features(
        dataframe=get_dataframe, rename_columns_dict={"Branche": "branch"}
    )
    result_df = feature_engineering(dataframe=df)

    assert isinstance(result_df, pd.DataFrame)
    assert isinstance(result_df.branch.dtype, type(np.dtype("int8")))
    assert len(result_df.branch.unique()) == 3
    assert all(result_df.branch.unique() != ["Offentligt", "Erhverv", "Privat"])
    assert all(result_df.branch.unique() == [1, 2, 3])



import datetime
from pathlib import Path

import pandas as pd
import pytest

from energy_consumption_forecasting.feature_pipeline.data_extraction import (
    extract_dataset_from_api,
    get_extraction_datetime,
)


@pytest.fixture
def date_range():
    start_date = datetime.datetime(2021, 1, 1, 12, 45)
    end_date = datetime.datetime(2021, 1, 10, 5, 10)
    return start_date, end_date


def test_valid_date_range_and_format(date_range):
    """
    In this test, both the start and end date are accurate.
    """
    result = get_extraction_datetime(date_range[0], date_range[1])

    assert isinstance(result, tuple)
    assert isinstance(result[0], str)
    assert isinstance(result[1], str)
    assert len(result) == 2
    assert result[0] < result[1]

    expected_format = "%Y-%m-%dT%H:%M"
    assert result[0] == date_range[0].strftime(expected_format)
    assert result[1] == (date_range[1] + datetime.timedelta(days=1)).strftime(
        expected_format
    )


def test_invalid_date_range(date_range):
    """
    In this test, the end date is smaller than the start date.
    """
    with pytest.raises(Exception) as exe_info:
        get_extraction_datetime(date_range[1], date_range[0])

    assert "End date needs to be greater than the start date" in str(exe_info.value)

    with pytest.raises(Exception) as exe_info:
        extract_dataset_from_api(date_range[1], date_range[0])

    assert "End date needs to be greater than the start date" in str(exe_info.value)


@pytest.mark.parametrize("save_data", [(False), (True)])
def test_extract_dataset_from_api(date_range, save_data):
    """
    Testing the function "extract_dataset_from_api()" whether it returns data from an
    API and save it in the directory.
    """

    result = extract_dataset_from_api(
        start_date_time=date_range[0],
        end_date_time=date_range[1],
        save_dataset_metadata=save_data,
    )
    if save_data:
        assert len(result) == 4
        assert isinstance(result[2], Path)
        assert isinstance(result[3], Path)
        assert result[2].is_file() == True
        assert result[3].is_file() == True

        # Deleting the generated files
        result[2].unlink()
        result[3].unlink()

    else:
        assert len(result) == 2
        assert isinstance(result, tuple)
        assert isinstance(result[0], pd.DataFrame)
        assert isinstance(result[1], dict)



import argparse
import datetime
import json
from pathlib import Path
from typing import Any, Dict, List, Optional, Sequence, Tuple

from pydantic import validate_call

from energy_consumption_forecasting.exceptions import log_exception
from energy_consumption_forecasting.feature_pipeline import (
    data_extraction,
    data_loading,
    data_transformation,
    data_validation,
)
from energy_consumption_forecasting.logger import get_logger
from energy_consumption_forecasting.utils import get_env_var, save_json_data

logger = get_logger(name=Path(__file__).name)
ROOT_DIRPATH = Path(get_env_var(key="PROJECT_ROOT_DIR_PATH", default_value="."))
DATA_DIRPATH = ROOT_DIRPATH / "data" / "processed_data"


@log_exception(logger=logger)
@validate_call
def run_feature_pipeline(
    start_date_time: datetime.datetime,
    end_date_time: datetime.datetime,
    rename_features: Dict[Any, Any],
    drop_features: Optional[List[Any]] = ["HourUTC"],
    check_features_duplicates: List[Any] = [
        "municipality_num",
        "branch",
        "datetime_dk",
    ],
) -> Tuple[Dict[Any, Any], Path]:
    """
    This functions runs the feature pipeline process i.e. ETL - Extract, Transform and
    Load.\n
    Data is extracted using the API, transformation is been done using the pandas
    library and data is been validated and loaded using great expectation and hopsworks.

    Parameters
    ----------
    start_date_time: datetime.datetime
        A starting date and time for extracting the data in datatype of
        datetime.datetime.

    end_date_time: datetime.datetime
        A ending date and time for extracting the data in datatype of
        datetime.datetime.

    rename_features: Dict[Any, Any]
        A dict containing existing column names as keys and new column names as values.

    drop_features: Optional[List[Any]], default=["HourUTC"]
        A list containing column names for dropping it from the DataFrame.

    check_features_duplicate: List[Any], default=["municipality_num", "branch", "datetime_dk"]
        A list containing column names for checking duplication and dropping the rows.

    Returns
    -------
    dict and pathlib.Path
        Returns a json metadata of the feature store in dict format and filepath of the
        JSON file.
    """

    # Extracting the dataset
    logger.info("Starting dataset extraction process.")

    dataframe, _, _, _ = data_extraction.extract_dataset_from_api(
        start_date_time=start_date_time,
        end_date_time=end_date_time,
    )

    logger.info("Data extraction process is successfully completed.\n")

    # Transforming the dataframe
    logger.info("Starting dataframe transformation process.")

    dataframe = data_transformation.rename_features(
        dataframe=dataframe,
        rename_columns_dict=rename_features,
    )
    dataframe = data_transformation.casting_features(dataframe=dataframe)
    dataframe = data_transformation.feature_engineering(dataframe=dataframe)
    dataframe = data_transformation.clean_dataframe(
        dataframe=dataframe,
        drop_columns=drop_features,
        check_columns_duplicates=check_features_duplicates,
    )

    logger.info("Data transformation process is successfully completed.\n")

    # Building the dataframe validation using the great expectation suite
    logger.info(
        "Starting the process to build the dataframe validation expectation suite."
    )

    generated_expectation_suite = data_validation.generate_great_expectation_suite()

    logger.info(
        "DataFrame validation expectation suite is been successfully generated.\n"
    )

    # Loading the dataframe into the feature store
    logger.info("Starting dataframe loading process.")

    feature_store_metadata, csv_filepath = data_loading.loading_data_to_hopsworks(
        dataframe=dataframe,
        generated_expectation_suite=generated_expectation_suite,
    )

    logger.info("Data loading process is successfully completed.\n")

    # Getting and cleaning the metadata and converting it into a JSON format
    feature_store_metadata = feature_store_metadata.json()
    feature_store_metadata = json.loads(
        feature_store_metadata.replace('\\"', '"')
        .replace('\\\\"', '"')
        .replace('"{', "{")
        .replace('}"', "}")
    )

    # Adding data extraction start and end datetime in metadata
    feature_store_metadata["data_extraction_start_datetime"] = str(start_date_time)
    feature_store_metadata["data_extraction_end_datetime"] = str(
        end_date_time + datetime.timedelta(days=1)
    )

    # Saving the provided feature store metadata in a local directory as a json file
    csv_filepath = csv_filepath.name.split("_")[:4]
    json_filepath = DATA_DIRPATH / f"{'_'.join(csv_filepath)}_metadata.json"
    save_json_data(data=feature_store_metadata, filepath=json_filepath)

    logger.info("Feature pipeline process is completed.")

    return feature_store_metadata, json_filepath


if __name__ == "__main__":

    class ParseKVAction(argparse.Action):

        @log_exception(logger=logger)
        def __call__(
            self,
            parser: argparse.ArgumentParser,
            namespace: argparse.Namespace,
            values: str | Sequence[Any] | None,
            option_string: str | None = None,
        ) -> None:
            setattr(namespace, self.dest, dict())
            for i in values:
                try:
                    key, value = i.split("=")
                    getattr(namespace, self.dest)[key] = value
                except ValueError as e:
                    message = f"\nTraceback: {e}"
                    message += f"\nError on {e}, format should be in 'key=value."
                    raise argparse.ArgumentError(self, str(message))

    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-s",
        "--start_datetime",
        type=datetime.datetime.fromisoformat,
        required=True,
        help="Starting date for extraction in format: YYYY-MM-DDTHH:MM:SS",
    )

    parser.add_argument(
        "-e",
        "--end_datetime",
        type=datetime.datetime.fromisoformat,
        required=True,
        help="Ending date for extraction in format: YYYY-MM-DDTHH:MM:SS",
    )

    parser.add_argument(
        "-r",
        "--rename_features",
        nargs="+",
        action=ParseKVAction,
        default={
            "HourDK": "datetime_dk",
            "MunicipalityNo": "municipality_num",
            "Branche": "branch",
            "ConsumptionkWh": "consumption_kwh",
        },
        help="Rename the features to transform the dataframe, "
        "format needs to be in KEY=VALUE pair, eg. CURRENT_NAME=NEW_NAME ...",
        metavar="CURRENT_NAME=NEW_NAME",
    )

    parser.add_argument(
        "-d",
        "--drop_features",
        nargs="+",
        default=["HourUTC"],
        help="Drop the features to transform the dataframe, "
        "format needs to be in NAME_1 NAME_2 NAME_3 ...",
    )

    parser.add_argument(
        "-c",
        "--check_duplicates",
        nargs="+",
        default=["municipality_num", "branch", "datetime_dk"],
        help="Provide feature names to check for duplicate data and "
        "clean the dataframe, format needs to be in NAME_1 NAME_2 NAME_3 ...",
    )

    args = parser.parse_args()

    _, filepath = run_feature_pipeline(
        start_date_time=args.start_datetime,
        end_date_time=args.end_datetime,
        rename_features=args.rename_features,
        drop_features=args.drop_features,
        check_features_duplicates=args.check_duplicates,
    )

    print(f"\nLocally saved metadata filepath: {filepath}")



from datetime import datetime

from airflow.decorators import dag, task
from airflow.models import Variable
from airflow.operators.empty import EmptyOperator
from airflow.utils.edgemodifier import Label
from airflow.utils.trigger_rule import TriggerRule


@dag(
    dag_id="workflow_pipeline",
    schedule="@daily",
    start_date=datetime(2023, 1, 1),
    catchup=True,
    tags=["feature-pipeline", "training-pipeline", "inference-pipeline"],
    max_active_runs=1,
)
def workflow_pipeline():
    """ """

    @task.virtualenv(
        task_id="run_feature_pipeline",
        requirements=[
            "--trusted-host 172.17.0.1:8000",
            "--index-url http://172.17.0.1:8000",
            "energy_consumption_forecasting",
        ],
        python_version="3.11.7",
        multiple_outputs=True,
        system_site_packages=True,
    )
    def run_feature_pipeline(
        start_date_time: str,
        end_date_time: str,
        rename_features: Dict[Any, Any],
        drop_features: Optional[List[Any]] = ["HourUTC"],
        check_features_duplicates: List[Any] = [
            "municipality_num",
            "branch",
            "datetime_dk",
        ],
    ):
        """ """



from great_expectations.core import ExpectationConfiguration, ExpectationSuite


def generate_great_expectation_suite() -> ExpectationSuite:
    """
    This functions generates an instance for Expectation Suite
    using the great expectation library.

    Returns
    -------
    great_expectation.core.ExpectationSuite
       A instance with all the predefined validation configuration for the Dataframe
       are been added in the suite using the ExpectationConfiguration module.
    """

    energy_suite = ExpectationSuite(
        expectation_suite_name="energy_consumption_forecast_suite"
    )

    # Adding DataFrame features in great expectation config for validation.
    # DataFrame Columns
    energy_suite.add_expectation(
        expectation_configuration=ExpectationConfiguration(
            expectation_type="expect_table_columns_to_match_ordered_list",
            kwargs={
                "column_list": [
                    "datetime_dk",
                    "municipality_num",
                    "branch",
                    "consumption_kwh",
                ]
            },
        )
    )

    energy_suite.add_expectation(
        expectation_configuration=ExpectationConfiguration(
            expectation_type="expect_table_column_count_to_equal", kwargs={"value": 4}
        )
    )

    # DataFrame Column: datetime_dk
    energy_suite.add_expectation(
        expectation_configuration=ExpectationConfiguration(
            expectation_type="expect_column_values_to_not_be_null",
            kwargs={"column": "datetime_dk"},
        )
    )

    # DataFrame Column: municipality_num
    energy_suite.add_expectation(
        expectation_configuration=ExpectationConfiguration(
            expectation_type="expect_column_values_to_not_be_null",
            kwargs={"column": "municipality_num"},
        )
    )

    energy_suite.add_expectation(
        expectation_configuration=ExpectationConfiguration(
            expectation_type="expect_column_values_to_be_of_type",
            kwargs={"column": "municipality_num", "type_": "int32"},
        )
    )

    energy_suite.add_expectation(
        expectation_configuration=ExpectationConfiguration(
            expectation_type="expect_column_distinct_values_to_be_in_set",
            kwargs={
                "column": "municipality_num",
                "value_set": (
                    250,
                    773,
                    766,
                    219,
                    326,
                    860,
                    420,
                    175,
                    820,
                    575,
                    240,
                    849,
                    573,
                    492,
                    665,
                    187,
                    482,
                    223,
                    329,
                    201,
                    410,
                    161,
                    756,
                    101,
                    540,
                    390,
                    813,
                    185,
                    306,
                    336,
                    147,
                    159,
                    450,
                    269,
                    340,
                    787,
                    707,
                    630,
                    710,
                    479,
                    265,
                    825,
                    173,
                    563,
                    360,
                    163,
                    791,
                    270,
                    740,
                    400,
                    217,
                    657,
                    607,
                    440,
                    253,
                    510,
                    151,
                    741,
                    779,
                    751,
                    376,
                    727,
                    350,
                    561,
                    330,
                    167,
                    840,
                    190,
                    480,
                    370,
                    210,
                    430,
                    760,
                    621,
                    316,
                    183,
                    706,
                    671,
                    810,
                    153,
                    580,
                    230,
                    169,
                    461,
                    165,
                    661,
                    155,
                    730,
                    157,
                    615,
                    260,
                    530,
                    746,
                    550,
                    846,
                    259,
                    851,
                    320,
                ),
            },
        )
    )

    # DataFrame Column: branch
    energy_suite.add_expectation(
        expectation_configuration=ExpectationConfiguration(
            expectation_type="expect_column_values_to_not_be_null",
            kwargs={"column": "branch"},
        )
    )

    energy_suite.add_expectation(
        expectation_configuration=ExpectationConfiguration(
            expectation_type="expect_column_values_to_be_of_type",
            kwargs={"column": "branch", "type_": "int8"},
        )
    )

    energy_suite.add_expectation(
        expectation_configuration=ExpectationConfiguration(
            expectation_type="expect_column_distinct_values_to_be_in_set",
            kwargs={"column": "branch", "value_set": (1, 2, 3)},
        )
    )

    # DataFrame Column: consumption_kwh
    energy_suite.add_expectation(
        expectation_configuration=ExpectationConfiguration(
            expectation_type="expect_column_values_to_not_be_null",
            kwargs={"column": "consumption_kwh"},
        )
    )

    energy_suite.add_expectation(
        expectation_configuration=ExpectationConfiguration(
            expectation_type="expect_column_values_to_be_of_type",
            kwargs={"column": "consumption_kwh", "type_": "float64"},
        )
    )

    energy_suite.add_expectation(
        expectation_configuration=ExpectationConfiguration(
            expectation_type="expect_column_min_to_be_between",
            kwargs={"column": "consumption_kwh", "min_value": 0},
        )
    )

    return energy_suite



from pathlib import Path
from typing import Tuple

import pandas as pd
from pydantic import validate_call
from sktime.split import temporal_train_test_split

import wandb
from energy_consumption_forecasting.exceptions import log_exception
from energy_consumption_forecasting.logger import get_logger
from energy_consumption_forecasting.training_pipeline.utils import init_wandb_run
from energy_consumption_forecasting.utils import get_env_var

logger = get_logger(name=Path(__file__).name)
ROOT_DIRPATH = Path(get_env_var(key="PROJECT_ROOT_DIR_PATH", default_value="."))

import hopsworks


def prepare_data(
    dataframe: pd.DataFrame,
    target_feature: str = "consumption_kwh",
    forecasting_horizon: int = 24,
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """
    This function prepares the dataset in train and test split for temporal dataframe.
    The test frame is based on the forecasting horizon.

    Parameters
    ----------
    dataframe: pd.DataFrame
        The dataframe that needs to be prepared.

    target_feature: str, default="consumption_kwh"
        The target feature in the dataframe.

    forecasting_horizon: int, default=24
        The forecasting horizon for the test split size in hours, by default 24 hours.

    Returns
    -------
        Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]
            Dataframe split in train and test: (y_train, y_test, X_train, X_test).
    """

    # Processing the data into sktime format:
    # X: Containing all the hierarchical features as multi-index format
    # y: Along with the multi-index features in X, target feature data needs to be present
    dataframe["datetime_dk"] = pd.PeriodIndex(dataframe["datetime_dk"], freq="H")
    dataframe = dataframe.set_index(
        ["municipality_num", "branch", "datetime_dk"]
    ).sort_index()

    # Splitting the input and target features
    X = dataframe.drop(columns=[target_feature])
    y = dataframe[[target_feature]]

    # Creating a time-series train test split
    y_train, y_test, X_train, X_test = temporal_train_test_split(
        y=y.sort_index(),
        X=X.sort_index(),
        test_size=forecasting_horizon,
    )

    return y_train, y_test, X_train, X_test


@log_exception(logger=logger)
@validate_call
def load_prepared_dataset_from_feature_store(
    feature_view_name: str = "denmark_energy_consumption_view",
    feature_view_ver: int = 1,
    training_dataset_ver: int = 1,
    target_feature: str = "consumption_kwh",
    forecasting_horizon: int = 24,
):
    """
    This function loads the feature view from feature store and gets the training
    dataset from the feature view.\nAll the metadata from feature view is stored in
    the WandB artifact with the same name of feature view.

    Once the data is received, the dataset is prepared (splitted in train and test) and
    the metadata for prepaid process is logged as artifact in Wandb with
    name "train_split" and "test_split".

    Parameters
    ----------
    feature_view_name: str, default="denmark_energy_consumption_view"
        The name of the feature view in the hopsworks feature store.

    feature_view_ver: int, default=1
        The feature view version that needs to be loaded.

    training_dataset_ver: int, default=1
        The training dataset version within the feature view that needs to be downloaded.

    target_feature: str, default="consumption_kwh"
        The name of the target feature in the dataset.

    forecasting_horizon: int, default=24
        The forecasting horizon for the test split size in hours, by default 24 hours.

    Returns
    -------
        Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]
            Dataframe splitted in train and test: (y_train, y_test, X_train, X_test).
    """

    # Connecting to the hopsworks feature store using the project API
    energy_project = hopsworks.login(
        project=get_env_var(key="FEATURE_STORE_PROJECT_NAME"),
        api_key_value=get_env_var(key="FEATURE_STORE_API_KEY"),
    )

    feature_store = energy_project.get_feature_store()

    logger.info(
        f'Connected to Hopsworks: Project Name "{energy_project.name}" and '
        f'Project URL: "{energy_project.get_url()}"'
    )

    # Getting the feature view metadata and training dataset from feature store
    # Creating a new artifact in wandb and storing the metadata in it.
    with init_wandb_run(
        run_name="get_training_dataset",
        job_type="load_feature_view",
        group="dataset",
    ) as run:

        logger.info(
            f"Details of the wandb run: run ID: {run.id}, run name: {run.name}, "
            f"project name: {run.project}, entity: {run.entity}, run group: "
            f"{run.group}, run save directory: {run.dir}, run URL: {run.url}"
        )

        feature_view = feature_store.get_feature_view(
            name=feature_view_name,
            version=feature_view_ver,
        )

        data, _ = feature_view.get_training_data(
            training_dataset_version=training_dataset_ver
        )

        metadata = feature_view.to_dict()
        metadata["query"] = metadata["query"].to_string()
        metadata["features"] = [i.name for i in metadata["features"]]
        metadata["link"] = feature_view._feature_view_engine._get_feature_view_url(
            feature_view=feature_view
        )
        metadata["feature_view_version"] = feature_view_ver
        metadata["training_dataset_version"] = training_dataset_ver

        run.log_artifact(
            wandb.Artifact(
                name=feature_view_name,
                type="feature_view",
                metadata=metadata,
            )
        )

        logger.info(f"Artifact {feature_view_name} has been logged successfully")

        run.finish()

    # Preparing the data by splitting it into train and test set
    # Logging the metadata for data preparation in wandb artifact.
    with init_wandb_run(
        run_name="train_test_split",
        job_type="prepare_dataset",
        group="dataset",
    ) as run:

        logger.info(
            f"Details of the wandb run: run ID: {run.id}, run name: {run.name}, "
            f"project name: {run.project}, entity: {run.entity}, run group: "
            f"{run.group}, run save directory: {run.dir}, run URL: {run.url}"
        )

        y_train, y_test, X_train, X_test = prepare_data(
            dataframe=data,
            target_feature=target_feature,
            forecasting_horizon=forecasting_horizon,
        )

        for i, (X, y) in enumerate(zip([X_train, X_test], [y_train, y_test])):

            metadata = {
                "dataset_time_interval": [
                    X.index.get_level_values(-1).min(),
                    X.index.get_level_values(-1).max(),
                ],
                "dataset_size": len(X),
                "total_municipality_num": len(X.index.get_level_values(0).unique()),
                "total_branch": len(X.index.get_level_values(1).unique()),
                "X_features": X.columns.to_list(),
                "y_features": y.columns.to_list(),
            }

            split = "train" if i == 0 else "test"
            run.log_artifact(
                wandb.Artifact(
                    name=f"{split}_split",
                    type="split_dataframe",
                    metadata=metadata,
                )
            )

            logger.info(f"Artifact {split}_split has been logged successfully")

        run.finish()

    return y_train, y_test, X_train, X_test



import logging
import os
from datetime import datetime
from pathlib import Path

from energy_consumption_forecasting.utils import get_env_var

ROOT_DIRPATH = Path(get_env_var(key="PROJECT_ROOT_DIR_PATH", default_value="."))
LOG_PATH = ROOT_DIRPATH / "logs"
FILE_NAME = f"{datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}.log"
LOG_FILENAME = LOG_PATH / FILE_NAME

if not os.path.isdir(LOG_PATH):
    os.makedirs(name=LOG_PATH)


def get_logger(name: str = "") -> logging.Logger:
    """
    This function provides a custom template for logging
    the details in the logs directory.

    Parameters
    ----------
    name: str, default = ""
        Name of the logger, eg. root or __main__

    Returns
    -------
    logger: logging.Logger
        returns a initiated logger object.
    """

    logging.basicConfig(
        filename=LOG_FILENAME,
        format=("[%(asctime)s] %(lineno)d - %(name)s - %(levelname)s - %(message)s"),
        level=logging.INFO,
    )

    logger = logging.getLogger(name=name)

    return logger



import argparse
import datetime
import json
from pathlib import Path
from typing import Any, Dict, Tuple

from pydantic import validate_call

from energy_consumption_forecasting.exceptions import (
    CustomExceptionMessage,
    log_exception,
)
from energy_consumption_forecasting.logger import get_logger
from energy_consumption_forecasting.utils import get_env_var, save_json_data

logger = get_logger(name=Path(__file__).name)
ROOT_DIRPATH = Path(get_env_var(key="PROJECT_ROOT_DIR_PATH", default_value="."))
DATA_DIRPATH = ROOT_DIRPATH / "data" / "processed_data"

import hopsworks


@log_exception(logger=logger)
@validate_call
def create_feature_view(
    start_datetime: datetime.datetime,
    end_datetime: datetime.datetime,
    feature_group_version: int = 1,
    feature_group_name: str = "denmark_energy_consumption_group",
    feature_views_name: str = "denmark_energy_consumption_view",
    feature_views_description: str = "Denmark's energy consumption forecasting model training view",
) -> Tuple[Dict[Any, Any], Path]:
    """
    This function creates a new feature view and training dataset on provided
    datetime and feature group data.

    Parameters
    ----------
    start_datetime: datetime.datetime
        A starting date and time for extracting the dataframe from feature group
        in datatype of datetime.datetime.

    end_datetime: datetime.datetime
        A ending date and time for extracting the dataframe from feature group
        in datatype of datetime.datetime.

    feature_group_version: int, default=1
         A version number in int type for getting the specific feature group.

    feature_group_name: str, default="denmark_energy_consumption_group"
        A feature group name in string type for getting the specific feature group.

    feature_views_name: str, default="denmark_energy_consumption_view"
        A feature view name in string type for naming the newly created view.

    feature_views_description: str, default="Denmark's energy consumption forecasting model training view"
        A feature view description in string format for setting the description of the
        newly created view.

    Returns
    -------
    Dict and pathlib.Path
        After creating the feature view and saving the metadata as a JSON file in the
        local directory, the function returns the metadata as a Dict object and the
        file path of the json file.
    """

    # Connecting to the hopsworks feature store using the project API
    energy_project = hopsworks.login(
        project=get_env_var(key="FEATURE_STORE_PROJECT_NAME"),
        api_key_value=get_env_var(key="FEATURE_STORE_API_KEY"),
    )

    feature_store = energy_project.get_feature_store()

    logger.info(
        f'Connected to Hopsworks: Project Name "{energy_project.name}" and '
        f'Project URL: "{energy_project.get_url()}"'
    )

    # Deleting old feature views because currently using free tier service of hopsworks
    # In free tier there is a limited options for creating views so replacing every time
    # a new view needs to be created.
    try:
        energy_feature_views = feature_store.get_feature_views(name=feature_views_name)
    except Exception as e:
        print(CustomExceptionMessage(exception_msg=e))
        logger.info(
            "Feature store could not get the feature view: "
            f'"{feature_views_name}" in the feature store.'
        )

        energy_feature_views = []

    if len(energy_feature_views) > 0:
        for view in energy_feature_views:
            try:
                view.delete_all_training_datasets()
            except Exception as e:
                print(CustomExceptionMessage(e))

            try:
                view.delete()
            except Exception as e:
                print(CustomExceptionMessage(e))

            logger.info(
                f'Deleting feature view for name "{feature_views_name}": {view}'
            )

    # Creating feature view from the provided feature group
    energy_feature_group = feature_store.get_feature_group(
        name=feature_group_name,
        version=feature_group_version,
    )

    dataframe_query = energy_feature_group.select_all()

    energy_feature_view = feature_store.create_feature_view(
        name=feature_views_name,
        query=dataframe_query,
        description=feature_views_description,
    )

    # Creating a training dataset in the feature views
    logger.info(
        f'Creating a training dataset in the feature view: "{feature_views_name}" '
        f"between the start date: {start_datetime} and end date: {end_datetime}"
    )

    energy_feature_view.create_training_data(
        description=f"Training dataset between {start_datetime} and {end_datetime}",
        data_format="csv",
        start_time=start_datetime,
        end_time=end_datetime,
        write_options={"wait_for_job": True},
        coalesce=False,
    )

    # Saving the metadata generated while creating the feature view
    energy_feature_view_metadata = energy_feature_view.json()
    energy_feature_view_metadata = json.loads(
        energy_feature_view_metadata.replace('\\"', '"')
        .replace('\\\\"', '"')
        .replace('"{', "{")
        .replace('}"', "}")
    )

    json_filepath = DATA_DIRPATH / f"{feature_views_name}_v1_metadata.json"
    save_json_data(data=energy_feature_view_metadata, filepath=json_filepath)

    logger.info(
        f"Feature view: '{feature_views_name}' and training dataset is been created and "
        f"the metadata is saved in the local directory as a JSON file: {json_filepath}."
    )

    return energy_feature_view_metadata, json_filepath


if __name__ == "__main__":

    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-s",
        "--start_datetime",
        type=datetime.datetime.fromisoformat,
        required=True,
        help="Starting date for extraction in format: YYYY-MM-DDTHH:MM:SS",
    )

    parser.add_argument(
        "-e",
        "--end_datetime",
        type=datetime.datetime.fromisoformat,
        required=True,
        help="Ending date for extraction in format: YYYY-MM-DDTHH:MM:SS",
    )

    parser.add_argument(
        "--group_version",
        type=int,
        default=1,
        help="Feature group version, needs to be in integer format.",
    )

    parser.add_argument(
        "--group_name",
        type=str,
        default="denmark_energy_consumption_group",
        help="Feature group name, needs to be in string format.",
    )

    parser.add_argument(
        "--views_name",
        type=str,
        default="denmark_energy_consumption_view",
        help="Feature view name within the feature group, needs to be in string format.",
    )

    parser.add_argument(
        "--views_desc",
        type=str,
        default="Denmark's energy consumption forecasting model training view",
        help="Description for the feature view, needs to be in string format.",
    )

    args = parser.parse_args()

    _, filepath = create_feature_view(
        start_datetime=args.start_datetime,
        end_datetime=args.end_datetime,
        feature_group_version=args.group_version,
        feature_group_name=args.group_name,
        feature_views_name=args.views_name,
        feature_views_description=args.views_desc,
    )

    print(f"\nLocally saved feature view metadata, filepath: {filepath}")



import os
from pathlib import Path
from typing import Optional

import hopsworks
import pandas as pd
from great_expectations.core import ExpectationSuite
from hsfs.feature_group import FeatureGroup
from pydantic import validate_call

from energy_consumption_forecasting.exceptions import log_exception
from energy_consumption_forecasting.logger import get_logger
from energy_consumption_forecasting.utils import get_env_var

logger = get_logger(name=Path(__file__).name)
ROOT_DIRPATH = Path(get_env_var(key="PROJECT_ROOT_DIR_PATH", default_value="."))
PROCESSED_DATA_DIRPATH = ROOT_DIRPATH / "data" / "processed_data"


@log_exception(logger=logger)
@validate_call(config=dict(arbitrary_types_allowed=True))
def loading_data_to_hopsworks(
    dataframe: pd.DataFrame,
    generated_expectation_suite: ExpectationSuite,
    hopsworks_feature_group_name: str = "denmark_energy_consumption_group",
    hopsworks_feature_group_version: int = 1,
    save_data_offline_dirpath: Optional[str | Path] = PROCESSED_DATA_DIRPATH,
) -> FeatureGroup | Path:
    """
    This function loads the provided dataframe in the hopsworks feature store.

    Parameters
    ----------
    dataframe: pd.DataFrame
        The pandas dataframe that needs to be uploaded in the feature store.

    generated_expectation_suite: ExpectationSuite
        A instance of the ExpectationSuite with predefined configurations for
        performing data validation while uploading the data in the feature store.

    hopsworks_feature_group_name: str, default='denmark_energy_consumption_group'
        A name in string type to set the feature group name.

    hopsworks_feature_group_version: int, default=1
        A version number in int type to set the feature group version.

    save_data_offline_dir: str or Path or None, default='./data/processed_data/'
        A directory path to save the dataframe in the directory, filename will be auto
        generated using the project name and dataset starting and ending datetime,
        the saved dataset will be in a CSV file format.

    Returns
    -------
    hsfs.feature_group.FeatureGroup or pathlib.Path
        Returns the metadata of the created and updated feature group, including all the
        details of the dataframe.
        OR returns metadata and filepath of the dataframe saved in local directory, if
        save_data_offline_dirpath argument is been provided.
    """

    # Connecting to the hopsworks feature store using the project API
    energy_feature_project = hopsworks.login(
        project=get_env_var(key="FEATURE_STORE_PROJECT_NAME"),
        api_key_value=get_env_var(key="FEATURE_STORE_API_KEY"),
    )

    logger.info(
        f'Connected to Hopsworks: Project Name "{energy_feature_project.name}" and '
        f'Project URL: "{energy_feature_project.get_url()}"'
    )

    hopsworks_feature_store = energy_feature_project.get_feature_store()

    # Creating a feature group within the hopsworks project
    energy_feature_group = hopsworks_feature_store.get_or_create_feature_group(
        name=hopsworks_feature_group_name,
        version=hopsworks_feature_group_version,
        description=(
            "Denmark's energy consumption data for industry, private and "
            "public categories per municipality number, data is provided on hourly "
            "base with an approx 15 days of update frequency."
        ),
        online_enabled=False,
        primary_key=["municipality_num", "branch"],
        expectation_suite=generated_expectation_suite,
        event_time="datetime_dk",
    )

    # Uploading the data using the above created feature group
    energy_feature_group.insert(
        features=dataframe,
        overwrite=False,
        write_options={"wait_for_job": True},
    )

    # Updating features metadata for the dataframe in the features group
    features_metadata = [
        {
            "name": "datetime_dk",
            "description": (
                "A date and time (interval), shown in Danish time zone. "
                "The interval between data is of an hourly difference."
            ),
        },
        {
            "name": "municipality_num",
            "description": (
                "Each of the 98 Danish municipalities has a unique number, "
                "ranging from 101 Copenhagen to 860 Hj\u00f8rring."
            ),
        },
        {
            "name": "branch",
            "description": (
                "All measurement id's related to a CVR number are labeled as 2 - 'Erhverv'"
                "except those that belong to 'Office, admin, which are labeled 1 - "
                "'Offentligt'. All other measurement id's are labeled 3 - 'Privat'."
            ),
        },
        {
            "name": "consumption_kwh",
            "description": (
                "Total energy consumption in kilowatt-hour(kWh), "
                "the value should be greater than 0"
            ),
        },
    ]

    for data in features_metadata:
        energy_feature_group.update_feature_description(
            feature_name=data.get("name"),
            description=data.get("description"),
        )

    # Updating statistic configuration for the feature group
    energy_feature_group.statistics_config = {
        "enabled": True,
        "histograms": True,
        "correlations": True,
    }

    # Calling, updating and computing the feature group statistic
    energy_feature_group.update_statistics_config()
    energy_feature_group.compute_statistics()

    logger.info(
        "Data and Metadata has been uploaded to the feature store "
        f'group: "{energy_feature_group.name}"'
    )

    # Saving the dataframe in the local data directory
    if save_data_offline_dirpath is not None:

        # Converting the string into a Path datatype
        if isinstance(save_data_offline_dirpath, str):
            save_data_offline_dirpath = Path(save_data_offline_dirpath)

        # Checking whether the path directory exist or not
        if not save_data_offline_dirpath.is_dir():
            os.makedirs(save_data_offline_dirpath)

        # Getting the start and end date of the dataframe
        start_date = (
            str(dataframe.datetime_dk.iloc[0]).replace(" ", "T").replace(":", "-")
        )
        end_date = (
            str(dataframe.datetime_dk.iloc[-1]).replace(" ", "T").replace(":", "-")
        )

        filepath = save_data_offline_dirpath / (
            f"{energy_feature_group.name}_v{hopsworks_feature_group_version}_"
            f"{start_date}_{end_date}.csv"
        )

        dataframe.to_csv(path_or_buf=filepath, index=False)
        logger.info(f'Dataset has been saved in csv file at "{filepath}".')

        return energy_feature_group, filepath

    return energy_feature_group



import argparse
import json
import os
from pathlib import Path
from typing import Any, Dict, List, Literal, Tuple

import joblib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from pydantic import validate_call
from sktime.performance_metrics.forecasting import (
    mean_absolute_percentage_error,
    mean_squared_percentage_error,
)
from sktime.utils.plotting import plot_series

import wandb
from energy_consumption_forecasting.exceptions import log_exception
from energy_consumption_forecasting.logger import get_logger
from energy_consumption_forecasting.training_pipeline.data_preprocessing import (
    load_prepared_dataset_from_feature_store,
)
from energy_consumption_forecasting.training_pipeline.model_builder import (
    build_lightgbm_model,
    build_naive_forecast_model,
)
from energy_consumption_forecasting.training_pipeline.utils import init_wandb_run
from energy_consumption_forecasting.utils import get_env_var, save_json_data

logger = get_logger(name=Path(__file__).name)
ROOT_DIRPATH = Path(get_env_var(key="PROJECT_ROOT_DIR_PATH", default_value="."))
DATA_DIRPATH = ROOT_DIRPATH / "data" / "assets" / "training"

import hopsworks


def train_model(model, X_train: pd.DataFrame, y_train: pd.DataFrame, fh: int = 24):
    """
    Train the forecast model using the given training dataset and forecast horizon.

    Parameters
    ----------
    model
        A sktime forecasting model or a forecasting pipeline.

    X_train: pd.DataFrame
        A pandas dataframe for training the model not containing the target feature.

    y_train: pd.DataFrame
        A pandas dataframe for training the model containing the target feature.

    Return
    ------
    model
        The model is been trained and returned back for performing prediction.
    """

    fh = np.arange(fh) + 1
    model.fit(y=y_train, X=X_train, fh=fh)

    return model


def perform_forecast(
    model, X_dataframe: pd.DataFrame, fh: int | None = None
) -> pd.DataFrame:
    """
    This function performs forecast on the provided model. The model should already
    be trained and the data provided should be different from the trained dataframe.

    Parameters
    ----------

    model
        A sktime forecasting model or a forecasting pipeline.

    X_dataframe: pd.DataFrame
        A pandas dataframe for performing forecast using the model,
        data should not contain the target feature.

    fh: int or None, default=None
        The forecasting horizon encoding the time stamps to forecast at.
        Should not be passed if has already been passed in model training.
        If has not been passed while model training, must be passed, not optional

    Returns
    -------
    pd.DataFrame
        Returns the predicted or forecasted dataframe from the point of forecast horizon
        with the same index as fh and same type as the data provided.
    """

    if fh is not None:
        fh = np.arange(fh) + 1

    return model.predict(X=X_dataframe, fh=fh)


def evaluate_model(
    model,
    X_test: pd.DataFrame,
    y_test: pd.DataFrame,
    fh: int | None = None,
    save_plot: bool = True,
    model_name: str = "forecast_model",
    model_id_or_ver: int = 1,
) -> Tuple[pd.DataFrame, Dict[str, float | pd.DataFrame]]:
    """
    This function evaluates the model and calculates the error deviation using the MAPE
    and RMSPE performance metrics.

    Parameters
    ----------
    model
        A sktime forecasting model or a forecasting pipeline.

    X_test: pd.DataFrame
        A pandas dataframe for training the model not containing the target feature.

    y_train: pd.DataFrame
        A pandas dataframe for training the model containing the target feature.

    fh: int or None, default=None
        The forecasting horizon encoding the time stamps to forecast at.
        Should not be passed if has already been passed in model training.
        If has not been passed while model training, must be passed, not optional

    save_plot: bool, default=True
        Whether to save the plot locally and log it ti wandb.

    model_name: str, default="forecast_model"
        Model name to create a directory to save the plots locally

    model_id_or_ver: int, default=1
        A integer for model identification.

    Returns
    -------
    Tuple[pd.DataFrame, Dict[str, float | pd.DataFrame]]
        A prediction dataframe and dict containing the performance metrics result.
    """

    # Creating a directory to save plots
    save_dirpath = DATA_DIRPATH / f"{model_name}_{model_id_or_ver}"
    if not save_dirpath.exists():
        os.makedirs(save_dirpath)

    # Performing model
    y_pred = perform_forecast(model=model, X_dataframe=X_test, fh=fh)

    # Calculating quantitative performance metrics and adding it in a result dict.
    result_dict = {}
    mape = mean_absolute_percentage_error(y_true=y_test, y_pred=y_pred, symmetric=False)
    rmspe = mean_squared_percentage_error(
        y_true=y_test, y_pred=y_pred, square_root=True
    )
    result_dict["mape"] = mape
    result_dict["rmspe"] = rmspe

    # Grouping the data and evaluating for every branch in the hierarchy
    grouped_result_df = pd.DataFrame(
        columns=["municipality_num", "branch", "mape", "rmspe"]
    )
    y_test_group = y_test.reset_index(level=[0, 1]).groupby(
        ["municipality_num", "branch"]
    )
    y_pred_group = y_pred.reset_index(level=[0, 1]).groupby(
        ["municipality_num", "branch"]
    )

    for y_test_group, y_pred_group in zip(y_test_group, y_pred_group):
        (y_test_m_num, y_test_branch), y_test_data = y_test_group
        (y_pred_m_num, y_pred_branch), y_pred_data = y_pred_group

        assert (
            y_test_m_num == y_pred_m_num and y_test_branch == y_pred_branch
        ), "y_test and y_pred grouped data are not aligned"

        # Calculating quantitative performance metric on every branched data
        mape_group = mean_absolute_percentage_error(
            y_true=y_test_data, y_pred=y_pred_data, symmetric=False
        )
        rmspe_group = mean_squared_percentage_error(
            y_true=y_test_data, y_pred=y_pred_data, square_root=True
        )

        group_df = pd.DataFrame(
            {
                "municipality_num": [y_test_m_num],
                "branch": [y_test_branch],
                "mape": [mape_group],
                "rmspe": [rmspe_group],
            }
        )
        grouped_result_df = pd.concat([grouped_result_df, group_df], ignore_index=True)

        # Plotting actual and prediction results and saving the plot
        if save_plot:
            fig, axs = plot_series(
                y_test_data.consumption_kwh,
                y_pred_data.consumption_kwh,
                labels=["Actual - y_test", "Prediction - y_pred"],
                markers=["", ""],
            )
            fig.suptitle(
                f"Municipality Num: {y_test_m_num} - Branch: {y_test_branch}\n"
            )

            plot_filepath = (
                save_dirpath / f"pred_plot_{y_test_m_num}_{y_test_branch}.png"
            )
            plt.savefig(plot_filepath)
            plt.close(fig)

            wandb.log({plot_filepath.stem: wandb.Image(str(plot_filepath))})

    # Plotting and saving performance metrics result
    if save_plot:
        fig, axs = plt.subplots(2, 1, figsize=(25, 25))
        fig.suptitle(
            t="Quantitative Performance Metrics: MAPE and RMSPE\n\n",
            weight="bold",
            size=25,
        )

        for i, metrics in enumerate(["mape", "rmspe"]):
            sns.barplot(
                data=grouped_result_df,
                x="municipality_num",
                y=metrics,
                hue="branch",
                palette="Set1",
                ax=axs[i],
            )
            axs[i].set_xticks(axs[i].get_xticks())
            axs[i].set_xticklabels(
                grouped_result_df.municipality_num.unique(), rotation=90, size=12
            )
            axs[i].tick_params(axis="y", labelsize=12)
            axs[i].set_xlabel("Municipality Number\n", size=20)
            axs[i].set_ylabel(f"{metrics.upper()}", size=20)
            axs[i].legend(
                title="Branch",
                title_fontsize=15,
                fontsize=12,
            )
        plt.tight_layout()
        plot_filepath = save_dirpath / "metrics_result.png"
        plt.savefig(plot_filepath)
        plt.close(fig)

        wandb.log({plot_filepath.stem: wandb.Image(str(plot_filepath))})

        logger.info(
            f"All the plot are saved in the directory: {save_dirpath} "
            "and also logged in the WandB."
        )
        logger.info(
            "Prediction comparison plot are saved in format "
            '"pred_plot_<municipality_num>_<branch>.png" and metrics_result.png'
        )

    # Adding the municipality_num and branch grouped
    # error metrics dataframe in the result dict
    result_dict["grouped_result_df"] = grouped_result_df

    return y_pred, result_dict


def save_model_to_hopsworks(
    model_name: str,
    model_ver: int,
    model_filepath: Path,
    model_metrics: Dict[str, float],
) -> Dict[str, Any]:
    """
    Save the model within the hopsworks registry, by connecting to the hopsworks
    account, to connect the hopsworks account your project name and API key needs to be
    mentioned as a env variable.

    Parameters
    ----------
    model_name: str
        Name of the model in string format.

    model_ver: int
        Version of the model in integer format.

    model_filepath: pathlib.Path
        Locally saved model filepath in pathlib.Path format.

    model_metrics: Dict[str, float]
        Performance metrics to log with the model needs to be in dict format.

    Returns
    -------
    Dict[str, Any]
        Once the model is registered the hopsworks model registry generates a
        metadata, containing all the details in dict format.

    """
    # Login in into hopsworks and registering the model
    project = hopsworks.login(
        project=get_env_var(key="FEATURE_STORE_PROJECT_NAME"),
        api_key_value=get_env_var(key="FEATURE_STORE_API_KEY"),
    )
    logger.info(
        f'Connected to Hopsworks: Project Name "{project.name}" and '
        f'Project URL: "{project.get_url()}"'
    )

    # Uploading the model to the registry
    model_registry = project.get_model_registry()
    model_meta_obj = model_registry.python.create_model(
        name=model_name, version=model_ver, metrics=model_metrics
    )
    model_meta_obj.save(model_path=model_filepath)
    logger.info(
        f"Model: {model_name} version: {model_ver} is been registered in hopsworks"
    )

    return model_meta_obj.to_dict()


@log_exception(logger=logger)
@validate_call
def model_training_pipeline(
    fh: int = 24,
    summarize_period: List[int] = [24, 48, 72],
    feature_view_name: str = "denmark_energy_consumption_view",
    feature_view_ver: int = 1,
    training_dataset_ver: int = 1,
    target_feature: str = "consumption_kwh",
    save_plot: bool = True,
    model_type: Literal["naive", "lightgbm"] = "lightgbm",
    model_name: str = "forecast_model",
    model_id_or_ver: int = 1,
) -> Dict[str, Any]:
    """
    This function performs the model training pipeline, where it builds the specified
    model, trains the model and evaluates the model performance. Once the steps are
    completed the model is logged as a artifact in WandB and registered in the
    hopsworks model registry.

    Parameters
    ----------

    fh: int, default=24
        A period that indicates the forecast horizon while making prediction in Hours.

    summarize_period: List[int], default=[24, 48, 72]
        The period at which the window summarizer transformer will be applied and
        calculate the lag, mean and std.
        For eg. [24, 48, 72] indicate: lag of 72 period, mean and std of first
        24 period, then 48 period and last 72 period.
        Note: Period can be sort of duration, in the above example it is hours.

    feature_view_name: str, default="denmark_energy_consumption_view"
        The name of the feature view in the hopsworks feature store.

    feature_view_ver: int, default=1
        The feature view version that needs to be loaded.

    training_dataset_ver: int, default=1
        The training dataset version within the feature view that needs to be downloaded.

    target_feature: str, default="consumption_kwh"
        The name of the target feature in the dataset.

    save_plot: bool, default=True
        Save the generated plot locally and log it in WandB, takes in only boolean values.

    model_type: Literal["naive", "lightgbm"], default="lightgbm",
        Build, train and evaluate the model, currently supports only two models "naive" and
        "lightgbm" as model_type values.

    model_name: str, default="forecast_model"
        Name of the model for identifying and using the model to save it locally
        and log it on WandB and hopsworks.

    model_id_or_ver: int, default=1
        Version or ID of the model for identification purpose.

    Returns
    -------
    Dict[str, Any]
        A metadata dict is returned, this metadata dict is also been logged in the WandB
        and saved locally as a JSON file at the location: "./data/assets/training/
        <model_name>_<model_ver_or_id>/<model_name>_<model_ver_or_id>.json"
    """

    # Getting the training and testing dataset from feature store
    y_train, y_test, X_train, X_test = load_prepared_dataset_from_feature_store(
        feature_view_name=feature_view_name,
        feature_view_ver=feature_view_ver,
        training_dataset_ver=training_dataset_ver,
        target_feature=target_feature,
        forecasting_horizon=fh,
    )

    # Getting the datetime range of the training and testing dataset
    train_start_datetime = y_train.index.get_level_values("datetime_dk").min()
    train_end_datetime = y_train.index.get_level_values("datetime_dk").max()
    test_start_datetime = y_test.index.get_level_values("datetime_dk").min()
    test_end_datetime = y_test.index.get_level_values("datetime_dk").max()

    logger.info(
        f'Training dataset datetime range is from "{train_start_datetime}" to '
        f"{train_end_datetime}."
    )
    logger.info(
        f'Testing dataset datetime range is from "{test_start_datetime}" to '
        f"{test_end_datetime}."
    )

    with init_wandb_run(
        run_name=f"{model_name}_{model_id_or_ver}",
        add_timestamp_to_run_name=True,
        job_type="model_training",
        group="model",
    ) as run:

        # Building and training model as per the provided model argument
        if model_type == "naive":
            forecast_model = build_naive_forecast_model(seasonal_periodicity=fh)
            logger.info("Naive forecaster model is been build and ready for training")

        elif model_type == "lightgbm":
            # Getting the saved hyperparameter file from the wandb artifact
            model_params_artifact = run.use_artifact(
                artifact_or_name="best_config:latest",
                type="model",
            )
            artifact_dir = model_params_artifact.download()
            model_params_path = Path(artifact_dir) / "best_config.json"
            with open(model_params_path) as file:
                model_params = json.load(file)

            # Updating the WandB config with the hyperparameters
            run.config.update(model_params)

            forecast_model = build_lightgbm_model(
                summarize_period=summarize_period,
                model_params=model_params,
            )
            logger.info("LightGBM model is been build and ready for training")

        forecast_model = train_model(
            model=forecast_model, X_train=X_train, y_train=y_train, fh=fh
        )

        y_pred, metrics_dict = evaluate_model(
            model=forecast_model,
            X_test=X_test,
            y_test=y_test,
            fh=fh,
            save_plot=save_plot,
            model_name=model_name,
            model_id_or_ver=model_id_or_ver,
        )
        metrics_df = metrics_dict.pop("grouped_result_df")

        pred_start_datetime = y_pred.index.get_level_values("datetime_dk").min()
        pred_end_datetime = y_pred.index.get_level_values("datetime_dk").max()
        logger.info(
            f'Predicted dataframe datetime range is from "{pred_start_datetime}" to '
            f"{pred_end_datetime}."
        )
        for k, v in metrics_dict.items():
            logger.info(f"Baseline metric dict: {k}: {v}")
        wandb.log({"metrics": {f"{model_name}_{model_id_or_ver}": metrics_dict}})
        wandb.log(
            {
                f"metric/{model_type}/{model_name}_{model_id_or_ver}": wandb.Table(
                    dataframe=metrics_df
                )
            }
        )

        # Saving the model locally as a pickle file
        save_dirpath = DATA_DIRPATH / f"{model_name}_{model_id_or_ver}"
        if not save_dirpath.exists():
            os.makedirs(save_dirpath)
        save_model_filepath = save_dirpath / f"{model_name}_{model_id_or_ver}.pkl"
        joblib.dump(value=forecast_model, filename=save_model_filepath)
        logger.info(
            f"Model pickle file is been saved locally at: {save_model_filepath}"
        )

        # Creating a wandb artifact of model and logging it in wandb
        # and also saving the model in the hopsworks model registry
        metadata = {
            "experiment": {
                "fh": fh,
                "summarize_period": summarize_period,
                "feature_view_name": feature_view_name,
                "feature_view_ver": feature_view_ver,
                "training_dataset_ver": training_dataset_ver,
                "target_feature": target_feature,
                "model_type": model_type,
                "model_name": model_name,
                "model_id_or_ver": model_id_or_ver,
                "training_start_datetime": train_start_datetime.to_timestamp().isoformat(),
                "training_end_datetime": train_end_datetime.to_timestamp().isoformat(),
                "testing_start_datetime": test_start_datetime.to_timestamp().isoformat(),
                "testing_end_datetime": test_end_datetime.to_timestamp().isoformat(),
                "prediction_start_datetime": pred_start_datetime.to_timestamp().isoformat(),
                "prediction_end_datetime": pred_end_datetime.to_timestamp().isoformat(),
            },
            "result": {"metrics": metrics_dict},
        }

        metadata["model_registry"] = save_model_to_hopsworks(
            model_name=model_name,
            model_ver=model_id_or_ver,
            model_filepath=save_model_filepath,
            model_metrics=metrics_dict,
        )

        artifact = wandb.Artifact(
            name=f"{model_name}_{model_id_or_ver}", type="model", metadata=metadata
        )
        artifact.add_file(local_path=save_model_filepath)
        run.log_artifact(artifact)

        metadata_filepath = save_dirpath / f"{model_name}_{model_id_or_ver}.json"
        save_json_data(data=metadata, filepath=metadata_filepath)
        logger.info(
            f"Training metadata file is been saved locally at: {metadata_filepath}"
        )

        run.finish()

    return metadata


if __name__ == "__main__":

    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--fh",
        type=int,
        default=24,
        help="Forecasting horizon period, needs to be in integer format.",
    )

    parser.add_argument(
        "--summarize_period",
        nargs="+",
        default=[24, 48, 72],
        help="Period to summarize the target feature, "
        "format needs to be in integer and multiple values are accepted.",
    )

    parser.add_argument(
        "--views_name",
        type=str,
        default="denmark_energy_consumption_view",
        help="Name of feature view within the feature group, needs to be in string format.",
    )

    parser.add_argument(
        "--views_ver",
        type=int,
        default=1,
        help="Version of feature view within the feature group, "
        "needs to be in integer format.",
    )

    parser.add_argument(
        "--dataset_ver",
        type=int,
        default=1,
        help="Version of training dataset within the feature view, "
        "needs to be in integer format.",
    )

    parser.add_argument(
        "--target_feature",
        type=str,
        default="consumption_kwh",
        help="Name of target feature, needs to be in string format.",
    )

    parser.add_argument(
        "--save_plot",
        type=bool,
        default=True,
        help="Whether to save plot while training and evaluating the model on dataset, "
        'only boolean values accepted either "True" or "False"',
    )

    parser.add_argument(
        "--model_type",
        type=str.lower,
        default="lightgbm",
        choices=["naive", "lightgbm"],
        help="Which type of model needs to be build, trained and forecasted, "
        'currently accepts only "naive" and "lightgbm" models.',
    )

    parser.add_argument(
        "--model_name",
        type=str,
        default="forecast_model",
        help="Name of the model for this training experiment, needs to be in string format.",
    )

    parser.add_argument(
        "--model_ver",
        type=int,
        default=1,
        help="Model version or id for this training experiment, "
        "needs to be in integer format.",
    )

    args = parser.parse_args()

    model_training_pipeline(
        fh=args.fh,
        summarize_period=args.summarize_period,
        feature_view_name=args.views_name,
        feature_view_ver=args.views_ver,
        training_dataset_ver=args.dataset_ver,
        target_feature=args.target_feature,
        save_plot=args.save_plot,
        model_type=args.model_type,
        model_name=args.model_name,
        model_id_or_ver=args.model_ver,
    )



from typing import Optional

import pandas as pd
from google.cloud import storage

from energy_consumption_forecasting.utils import get_env_var

# Google cloud platform environment variables for project, bucket and service account
GCS_PROJECT = get_env_var(key="GOOGLE_CLOUD_PROJECT")
GCS_BUCKET_NAME = get_env_var(key="GOOGLE_CLOUD_BUCKET_NAME")
GCS_SERVICE_ACCOUNT_JSON_PATH = get_env_var(
    key="GOOGLE_CLOUD_SERVICE_ACCOUNT_JSON_PATH"
)


def get_gcs_bucket(
    project: str = GCS_PROJECT,
    bucket_name: str = GCS_BUCKET_NAME,
    json_credentials_path: str = GCS_SERVICE_ACCOUNT_JSON_PATH,
) -> storage.Bucket:
    """
    This function gets the Google cloud storage(GCS) bucket from GCP, using this bucket
    object you can upload and download data from the Google cloud storage.

    Parameters
    ----------
    project: str
        Connects to the GCS using the project name, by default uses the
        environment variable called GOOGLE_CLOUD_PROJECT.

    bucket_name: str
        Connects to the GCS using the bucket name, by default uses the
        environment variable called GOOGLE_CLOUD_BUCKET_NAME.

    json_credentials_path: str
        Connects to the GCS using the JSON filepath that was generated when an
        service account was created with roles to manipulate the bucket and objects
        in GCS, by default uses the environment variable called
        GCS_SERVICE_ACCOUNT_JSON_PATH.

    Returns
    -------
    storage.Bucket
        A bucket object received from Google cloud storage to upload and download data.
    """
    client = storage.Client.from_service_account_json(
        json_credentials_path=json_credentials_path, project=project
    )

    bucket = client.bucket(bucket_name=bucket_name)

    return bucket


def write_blob_to_bucket(bucket: storage.Bucket, blob_name: str, data: pd.DataFrame):
    """
    This function gets the binary large object(blob) from the Google cloud storage(GCS)
    bucket. Using the blob object, dataframe can be stored as binary file for
    which the data is transformed into apache parquet.

    Parameters
    ----------
    bucket: storage.Bucket
        A bucket object that is generated by GCS, can be used for uploading and
        downloading data.

    blob_name: str
        The name of the blob to be instantiated.

    data: pd.DataFrame
        The data that needs to be stored in the GCS.
    """
    blob_obj = bucket.blob(blob_name=blob_name)

    with blob_obj.open(mode="wb") as file:
        data.to_parquet(path=file)


def read_blob_from_bucket(
    bucket: storage.Bucket, blob_name: str
) -> Optional[pd.DataFrame]:
    """
    This function gets the binary large object(blob) from the Google cloud storage(GCS)
    bucket. Using the blob object, the binary data is transformed back to pandas
    dataframe.

    Parameters
    ----------
    bucket: storage.Bucket
        A bucket object that is generated by GCS, can be used for uploading and
        downloading data.

    blob_name: str
        The name of the blob to be instantiated.

    Returns
    -------
    pd.DataFrame or None
        A dataframe containing the data within the blob or None if the blob does not
        exists.
    """
    blob_obj = bucket.blob(blob_name=blob_name)

    if not blob_obj.exists():
        return None

    with blob_obj.open("rb") as file:
        return pd.read_parquet(path=file)



import argparse
from datetime import datetime
from pathlib import Path

import joblib
import numpy as np
import pandas as pd
from pydantic import validate_call

from energy_consumption_forecasting.exceptions import log_exception
from energy_consumption_forecasting.inference_pipeline.batch_data import (
    get_batch_data_from_hopsworks,
)
from energy_consumption_forecasting.inference_pipeline.utils import (
    get_gcs_bucket,
    read_blob_from_bucket,
    write_blob_to_bucket,
)
from energy_consumption_forecasting.logger import get_logger
from energy_consumption_forecasting.utils import get_env_var

logger = get_logger(name=Path(__file__).name)

import hopsworks


def load_model_from_hopsworks(
    project,
    model_version: int = 1,
    model_name: str = "forecast_model",
):
    """
    This function loads the model from hopsworks model registry,
    model is downloaded as pickle file and transformed into a
    sktime object that can be used for forecasting.

    Parameters
    ----------
    project
        A project object generated by hopsworks when login into a specific project.

    model_version: int, default=1
        Model version used when saving the model in hopsworks.

    model_name: str, default="forecast_model"
        Model name used when saving the model in hopsworks.

    Returns
    -------
    model
        A model is returned after loading the pickle file, the model type can be a
        sktime model or sktime forecasting pipeline which can be used for forecasting.
    """

    model_registry = project.get_model_registry()
    model_instance = model_registry.get_model(name=model_name, version=model_version)
    model_path = model_instance.download()
    model_path = Path(model_path) / f"{model_name}_{model_version}.pkl"

    # Loading the model pickle file object
    model = joblib.load(model_path)

    return model


def get_batch_forecast(model, X: pd.DataFrame, fh: int = 24) -> pd.DataFrame:
    """
    This functions takes the X dataframe and generates a forecast dataframe similar to
    X with the help of forecast horizon input.
    Once the data is generated it is used to make prediction from the model.

    Parameters
    ----------
    model:
        A sktime model or sktime forecasting pipeline that has been trained using the
        fit function.

    X: pd.DataFrame
        A batch dataframe with input features that are provided by the feature store.

    fh: int, default=24
        A period that indicates the forecast horizon while making prediction in Hours.

    Returns
    -------
    pd.Dataframe
        Model generates the forecast prediction in pandas dataframe format.
        The range of the forecast date and time depends on the fh argument.
    """

    # Getting the index exogenous features and creating a forecast date range
    municipality_num = X.index.get_level_values(level=0).unique()
    branch = X.index.get_level_values(level=1).unique()
    last_datetime = X.index.get_level_values(level=2).max()

    start_datetime = last_datetime + 1
    end_datetime = last_datetime + fh
    fh_range = pd.date_range(
        start=start_datetime.to_timestamp(),
        end=end_datetime.to_timestamp(),
        freq="H",
    )
    fh_range = pd.PeriodIndex(data=fh_range, freq="H")

    # Creating a forecast X data using the above generated data
    multi_index = pd.MultiIndex.from_product(
        iterables=[municipality_num, branch, fh_range],
        names=["municipality_num", "branch", "datetime_dk"],
    )
    X_forecast = pd.DataFrame(index=multi_index)

    # Performing prediction using the newly generated forecast data and forecast horizon
    prediction = model.predict(X=X_forecast, fh=np.arange(fh) + 1)

    return prediction


def save_data_to_gcs_bucket(X: pd.DataFrame, y: pd.DataFrame, prediction: pd.DataFrame):
    """
    This function saves the data into the Google cloud storage bucket by transforming
    the provided dataframe into parquet data and uploading it as blob object in
    the bucket.

    Parameters
    ----------
    X: pd.DataFrame
        A dataframe that is used as input data for generating prediction from model.
        This data is saved in the bucket with the blob name as "input.parquet".

    y: pd.DataFrame
        A dataframe that is used as target data for generating prediction from model.
        This data is saved in the bucket with the blob name as "target.parquet".

    prediction: pd.DataFrame
        A dataframe that is generated from the model as a prediction or forecast.
        This data is saved in the bucket with the blob name as "prediction.parquet".
    """

    # Connecting to Google cloud storage and getting the bucket object
    bucket = get_gcs_bucket()

    # Transforming and uploading the input, target and prediction dataframe
    # into parquet data in the GCS bucket
    for data, blob_name in zip(
        [X, y, prediction], ["input.parquet", "target.parquet", "prediction.parquet"]
    ):
        logger.info(f'Uploading blob data "{blob_name}" in the GCS bucket.')
        write_blob_to_bucket(
            bucket=bucket,
            blob_name=blob_name,
            data=data,
        )
        logger.info(f'Blob Data "{blob_name}" is successfully uploaded to GCS bucket.')


def save_prediction_data_for_caching(
    prediction: pd.DataFrame,
    start_datetime: datetime,
):
    """
    This function creates or updates the cached prediction dataframe with the newly
    generated prediction and then saves the dataframe in the Google cloud storage(GCS)
    as a parquet file called "cached_prediction.parquet".

    Parameters
    ----------
    prediction: pd.DataFrame
        Newly generated forecast prediction by the model.

    start_datetime: datetime.datetime
        A starting date and time for filtering the cached prediction before uploading
        the data in the GCS bucket.
    """

    # Getting the GCS bucket object to download the cached prediction data
    bucket = get_gcs_bucket()
    cached_prediction = read_blob_from_bucket(
        bucket=bucket, blob_name="cached_prediction.parquet"
    )

    # Checking if cached prediction data exist
    cached_prediction_exist = cached_prediction is not None
    if cached_prediction_exist:
        # Merging current prediction and cached prediction
        merged_prediction = prediction.merge(
            right=cached_prediction,
            how="outer",
            left_index=True,
            right_index=True,
            suffixes=("_new", "_cached"),
        )

        # Filtering the new and cached prediction into different dataframe
        new_prediction = merged_prediction.filter(regex="_new")
        new_prediction.columns = new_prediction.columns.str.replace(pat="_new", repl="")
        cached_pred = merged_prediction.filter(regex="_cached")
        cached_pred.columns = cached_pred.columns.str.replace(pat="_cached", repl="")

        # Updating the NaN values in the new prediction with cached prediction data
        new_prediction.update(cached_pred)
        prediction = new_prediction

    # Checking the prediction dataframe and uploading it to the GCS bucket
    prediction = prediction.loc[
        prediction.index.get_level_values("datetime_dk")
        >= pd.Period(start_datetime, freq="H")
    ]
    prediction = prediction.dropna(subset=["consumption_kwh"])

    write_blob_to_bucket(
        bucket=bucket,
        blob_name="cached_prediction.parquet",
        data=prediction,
    )
    logger.info(
        'Cached prediction "cached_prediction.parquet" has been successfully '
        "updated with the new forecast predictions"
    )


@log_exception(logger=logger)
@validate_call
def run_inference_pipeline(
    start_datetime: datetime,
    end_datetime: datetime,
    fh: int = 24,
    target_feature: str = "consumption_kwh",
    feature_view_ver: int = 1,
    feature_view_name: str = "denmark_energy_consumption_view",
    model_version: int = 1,
    model_name: str = "forecast_model",
):
    """
    This function will perform the batch inference pipeline from getting the data
    and model from hopsworks and using the model to get the forecast prediction and
    finally storing the data in the Google cloud storage(GCS) bucket and also
    caching the prediction for future use.

    Parameters
    ----------

    start_datetime: datetime.datetime
        A starting date and time for extracting the batch data
        from hopsworks feature view.

    end_datetime: datetime.datetime
        A ending date and time for extracting the batch data
        from hopsworks feature view.

    fh: int, default=24
        A period that indicates the forecast horizon while making prediction in Hours.

    target_feature: str, default="consumption_kwh"
        The name of the target feature in the dataset.

    feature_view_ver: int, default=1
        The feature view version that needs to be loaded.

    feature_view_name: str, default="denmark_energy_consumption_view"
        The name of the feature view in the hopsworks feature store.

    model_version: int, default=1
        Model version used when saving the model in hopsworks.

    model_name: str, default="forecast_model"
        Model name used when saving the model in hopsworks.
    """

    # Connecting to the hopsworks feature store using the project API
    project = hopsworks.login(
        project=get_env_var(key="FEATURE_STORE_PROJECT_NAME"),
        api_key_value=get_env_var(key="FEATURE_STORE_API_KEY"),
    )
    feature_store = project.get_feature_store()
    logger.info(
        f'Connected to Hopsworks: Project Name "{project.name}" and '
        f'Project URL: "{project.get_url()}"'
    )

    # Getting the dataframe from the hopsworks feature store and transforming for inference
    logger.info("Loading data from hopsworks feature store.")
    logger.info(
        "Loading data between the following date-time range, "
        f"start: {start_datetime} and end: {end_datetime}"
    )
    X, y = get_batch_data_from_hopsworks(
        feature_store=feature_store,
        start_datetime=start_datetime,
        end_datetime=end_datetime,
        feature_view_ver=feature_view_ver,
        feature_view_name=feature_view_name,
        target_feature=target_feature,
    )
    logger.info("Data is successfully loaded from the hopsworks feature store.")

    # Getting the model from the hopsworks model registry
    logger.info("Loading model from the hopsworks model registry.")
    model = load_model_from_hopsworks(
        project=project,
        model_version=model_version,
        model_name=model_name,
    )
    logger.info("Model is successfully loaded from the model registry.")

    # Performing forecast prediction using the data and model
    logger.info("Performing forecast prediction using the data and model.")
    prediction = get_batch_forecast(model=model, X=X, fh=fh)
    pred_start_datetime = prediction.index.get_level_values(level="datetime_dk").min()
    pred_end_datetime = prediction.index.get_level_values(level="datetime_dk").max()
    logger.info(
        f"Forecast prediction are from {pred_start_datetime} to {pred_end_datetime}"
    )
    logger.info("Prediction were successfully forecasted.")

    # Saving the dataframe in the GCS bucket
    logger.info("Saving the dataframe in the GCS bucket.")
    save_data_to_gcs_bucket(
        X=X,
        y=y,
        prediction=prediction,
    )
    logger.info("All the dataframe were successfully uploaded in the GCS bucket.")

    # Saving the newly generated prediction by updating the cached_prediction dataframe
    logger.info("Merging new prediction with cached prediction")
    save_prediction_data_for_caching(
        prediction=prediction,
        start_datetime=start_datetime,
    )
    logger.info("Cached prediction was successfully updated with the new prediction")


if __name__ == "__main__":

    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-s",
        "--start_datetime",
        type=datetime.fromisoformat,
        required=True,
        help="Starting date for extraction in format: YYYY-MM-DDTHH:MM:SS",
    )

    parser.add_argument(
        "-e",
        "--end_datetime",
        type=datetime.fromisoformat,
        required=True,
        help="Ending date for extraction in format: YYYY-MM-DDTHH:MM:SS",
    )

    parser.add_argument(
        "--fh",
        type=int,
        default=24,
        help="Forecasting horizon period, needs to be in integer format.",
    )

    parser.add_argument(
        "--views_name",
        type=str,
        default="denmark_energy_consumption_view",
        help="Name of feature view within the feature group, needs to be in string format.",
    )

    parser.add_argument(
        "--views_ver",
        type=int,
        default=1,
        help="Version of feature view within the feature group, "
        "needs to be in integer format.",
    )

    parser.add_argument(
        "--target_feature",
        type=str,
        default="consumption_kwh",
        help="Name of target feature, needs to be in string format.",
    )

    parser.add_argument(
        "--model_name",
        type=str,
        default="forecast_model",
        help="Name of the model for this training experiment, needs to be in string format.",
    )

    parser.add_argument(
        "--model_ver",
        type=int,
        default=1,
        help="Model version or id for this training experiment, "
        "needs to be in integer format.",
    )

    args = parser.parse_args()

    run_inference_pipeline(
        start_datetime=args.start_datetime,
        end_datetime=args.end_datetime,
        fh=args.fh,
        target_feature=args.target_feature,
        feature_view_ver=args.views_ver,
        feature_view_name=args.views_name,
        model_version=args.model_ver,
        model_name=args.model_name,
    )



import logging
import sys
import traceback
from functools import wraps


class CustomExceptionMessage(Exception):
    """
    This class provides custom exception message.

    Parameters
    ----------
    error: Exception
        Provide the error that will be generated by the Exception class.

    Returns
    -------
    error_message: str
        A string containing the script file name, error line number and the
        exception message.
    """

    def __init__(self, exception_msg: Exception):
        super().__init__(exception_msg)
        self.exception_msg = exception_msg
        self.error_message = self.error_message_detail()

    def __str__(self) -> str:
        return self.error_message

    def error_message_detail(self) -> str:
        _, _, exc_traceback = sys.exc_info()
        file_name = exc_traceback.tb_frame.f_code.co_filename
        line_number = exc_traceback.tb_lineno

        traceback.print_exc()
        error_message = (
            "\n[Exception] A exception occurred in python module: "
            f"[{file_name}] at line number: [{line_number}] with error "
            f"message:\n {type(self.exception_msg).__name__} - {self.exception_msg}"
        )

        return error_message


def log_exception(logger: logging.Logger):
    """
    A decorator for catching exception and logging them in the log file.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except:
                issue = "exception in " + func.__name__ + "\n"
                issue = issue + "==========================\n"
                logger.exception(issue)
                raise

        return wrapper

    return decorator


import json
import os
from pathlib import Path
from typing import Optional

from dotenv import load_dotenv


def get_env_var(
    key: str,
    default_value: str = None,
    env_path: Optional[str | Path] = None,
) -> Optional[str]:
    """
    Loads the environment variables and returns the value for the requested key.

    Parameters
    ----------
    key: str
        The key of the requested environment variable.

    default_value: str, default=None
        The value to be returned if environment variable is not found.

    env_path: str or Path or None, default=None
        A path to the .env file.

    Returns
    -------
    value: str or None
        The value of the environment variable that is requested.
    """
    if isinstance(env_path, str):
        env_path = Path(env_path)

    load_dotenv(
        dotenv_path=env_path,
        override=True,
    )

    if default_value is None:
        value = os.getenv(key=key)
    else:
        value = os.getenv(key=key, default=default_value)

    return value


def save_json_data(data: dict, filepath: str | Path):
    """
    Saves a dictionary object as a JSON file.

    Parameters
    ----------
    data: dict
        The dict object that needs to saved as a JSON file.

    filepath: str or path
        A file path as a string or pathlib.Path object, to save the data at the location.
        Name of the file must be with the extension of ".json".
    """
    if isinstance(filepath, str):
        filepath = Path(filepath)

    if not filepath.suffix == ".json":
        raise Exception(
            f'Filepath needs to have a extension of .json: "{filepath}", '
            "but was not provided."
        )

    # Saving the data as a json file
    with open(file=filepath, mode="w", encoding="utf-8") as file:
        json.dump(data, file)



from datetime import datetime
from typing import Tuple

import pandas as pd
from hsfs.feature_store import FeatureStore


def get_batch_data_from_hopsworks(
    feature_store: FeatureStore,
    start_datetime: datetime,
    end_datetime: datetime,
    feature_view_ver: int = 1,
    feature_view_name: str = "denmark_energy_consumption_view",
    target_feature: str = "consumption_kwh",
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    This functions get the batch data from hopsworks feature view and process the batch
    data into the sktime data format for hierarchical dataset and then returns the
    input and target dataset.

    Parameters
    ----------
    feature_store: FeatureStore
        A instance of FeatureStore object that is connected to the hopsworks feature
        store and the project in which the data is stored.

    start_datetime: datetime.datetime
        A starting date and time for extracting the batch data
        from hopsworks feature view.

    end_datetime: datetime.datetime
        A ending date and time for extracting the batch data
        from hopsworks feature view.

    feature_view_ver: int, default=1
        The feature view version that needs to be loaded.

    feature_view_name: str, default="denmark_energy_consumption_view"
        The name of the feature view in the hopsworks feature store.

    target_feature: str, default="consumption_kwh"
        The name of the target feature in the dataset.

    Returns
    -------
    Tuple[pd.DataFrame, pd.DataFrame]
            Dataframe split in Input and Target sets as per sktime format: X, y.
    """

    # Getting the feature views where the data is been stored for batch processing
    feature_view = feature_store.get_feature_view(
        name=feature_view_name, version=feature_view_ver
    )

    batch_data = feature_view.get_batch_data(
        start_time=start_datetime, end_time=end_datetime
    )

    # Processing the data into sktime format:
    # X: Containing all the hierarchical features as multi-index format
    # y: Along with the multi-index features in X, target feature data needs to be present
    batch_data["datetime_dk"] = pd.PeriodIndex(batch_data["datetime_dk"], freq="H")
    batch_data = batch_data.set_index(
        ["municipality_num", "branch", "datetime_dk"]
    ).sort_index()

    # Splitting the input and target features
    X = batch_data.drop(columns=[target_feature])
    y = batch_data[[target_feature]]

    return X, y



from typing import Any, Dict, List, Optional

import lightgbm as lgb
from sktime.forecasting.compose import ForecastingPipeline, make_reduction
from sktime.forecasting.naive import NaiveForecaster
from sktime.transformations.series.date import DateTimeFeatures
from sktime.transformations.series.summarize import WindowSummarizer


def build_naive_forecast_model(seasonal_periodicity: int = 24) -> NaiveForecaster:
    """
    This function builds a naive forecast model using the 'last' as a strategy,
    where the forecast will be based on the last value of each season in the series.

    Parameters
    ----------
    seasonal_periodicity: int, default=1
        A integer for defining the seasonal period of the series forecasting.

    Returns
    -------
    sktime.forecasting.naive.NaiveForecaster
        Returns the NaiveForecaster model with the specified seasonal period.
    """
    return NaiveForecaster(sp=seasonal_periodicity)


def build_lightgbm_model(
    summarize_period: List[int] = [24, 48, 72],
    model_params: Optional[Dict[str, Any]] = None,
):
    """
    This function builds a LightGBM model using the sktime method and returns a
    forecasting pipeline which can be used to fit and predict the time series data.

    Parameters
    ----------
    summarize_period: List[int], default=[24, 48, 72]
        The period at which the window summarizer transformer will be applied and
        calculate the lag, mean and std.
        For eg. [24, 48, 72] indicate: lag of 72 period, mean and std of first
        24 period, then 48 period and last 72 period.
        Note: Period can be sort of duration, in the above example it is hours.

    model_params: Dict[str, Any] or None, default=None
        A dict containing the hyperparameters of the LightGBM model.
        If None is provided then model is build with default parameters.

    Returns
    -------
    sktime.forecasting.compose.ForecastingPipeline
        A sktime forecasting pipeline is returned containing the LightGBM model and
        all the necessary transformer, this can be used for training and predicting.
    """

    # Creating window summarizer to transform the dataframe
    kwargs = {
        "lag_feature": {
            "lag": list(range(1, summarize_period[-1] + 1)),
            "mean": [[1, i] for i in summarize_period],
            "std": [[1, i] for i in summarize_period],
        }
    }

    window_summarizer = WindowSummarizer(**kwargs)

    # Building the LightGBM model on sktime forecaster method
    if model_params is None:
        lgbm_regressor = lgb.LGBMRegressor()
    else:
        lgbm_regressor = lgb.LGBMRegressor(**model_params)

    forecaster = make_reduction(
        estimator=lgbm_regressor,
        transformers=[window_summarizer],
        window_length=None,
        strategy="recursive",
        pooling="global",
    )

    # Building the forecast pipeline for fitting the data in the model
    pipe = ForecastingPipeline(
        steps=[
            (
                "daily_season",
                DateTimeFeatures(
                    manual_selection=["day_of_week", "hour_of_day"],
                    keep_original_columns=True,
                ),
            ),
            ("forecaster", forecaster),
        ]
    )

    return pipe



import argparse
from pathlib import Path
from typing import Any, Dict, List, Optional

import numpy as np
import optuna
import pandas as pd
import wandb
from optuna.visualization import (
    plot_optimization_history,
    plot_param_importances,
    plot_slice,
)
from pydantic import validate_call
from sktime.performance_metrics.forecasting import mean_absolute_percentage_error

from energy_consumption_forecasting.exceptions import log_exception
from energy_consumption_forecasting.logger import get_logger
from energy_consumption_forecasting.training_pipeline.data_preprocessing import (
    load_prepared_dataset_from_feature_store,
)
from energy_consumption_forecasting.training_pipeline.model_builder import (
    build_lightgbm_model,
)
from energy_consumption_forecasting.training_pipeline.utils import init_wandb_run
from energy_consumption_forecasting.utils import get_env_var, save_json_data

logger = get_logger(name=Path(__file__).name)
ROOT_DIRPATH = Path(get_env_var(key="PROJECT_ROOT_DIR_PATH", default_value="."))

# Getting arguments for model tuning and hyperparameter tuning
parser = argparse.ArgumentParser()

parser.add_argument(
    "--views_name",
    type=str,
    default="denmark_energy_consumption_view",
    help="Name of feature view within the feature group, needs to be in string format.",
)

parser.add_argument(
    "--views_ver",
    type=int,
    default=1,
    help="Version of feature view within the feature group, "
    "needs to be in integer format.",
)

parser.add_argument(
    "--dataset_ver",
    type=int,
    default=1,
    help="Version of training dataset within the feature view, "
    "needs to be in integer format.",
)

parser.add_argument(
    "--target_feature",
    type=str,
    default="consumption_kwh",
    help="Name of target feature, needs to be in string format.",
)

parser.add_argument(
    "--fh",
    type=int,
    default=24,
    help="Forecasting horizon period, needs to be in integer format.",
)

parser.add_argument(
    "--summarize_period",
    nargs="+",
    default=[24, 48, 72],
    help="Period to summarize the target feature, "
    "format needs to be in integer and multiple values are accepted.",
)

parser.add_argument(
    "--n_trials",
    type=int,
    default=20,
    help="Number of trials to study and get the best hyperparameters, "
    "needs to be in integer format.",
)

parser.add_argument(
    "-f",
    "--filepath",
    type=Path,
    default=ROOT_DIRPATH / "data" / "assets" / "hyperparameter" / "best_config.json",
    help="Filepath to save the best hyperparameter config in local directory as a "
    "JSON file, argument needs to have a .json extension.",
)

args = parser.parse_args()

logger.info("Getting the dataset from feature store.")

# Getting the training and testing dataframe
y_train, y_test, X_train, X_test = load_prepared_dataset_from_feature_store(
    feature_view_name=args.views_name,
    feature_view_ver=args.views_ver,
    training_dataset_ver=args.dataset_ver,
    target_feature=args.target_feature,
    forecasting_horizon=args.fh,
)

logger.info("Train and test dataset is available.")


@log_exception(logger=logger)
@validate_call(config=dict(arbitrary_types_allowed=True))
def model_tuning(
    X_train: pd.DataFrame = X_train,
    y_train: pd.DataFrame = y_train,
    X_test: pd.DataFrame = X_test,
    y_test: pd.DataFrame = y_test,
    fh: int = 24,
    summarize_period: List[int] = [24, 48, 72],
    model_params: Optional[Dict[str, Any]] = None,
) -> float:
    """
    This function build the lightgbm model and trains and predict on training and
    testing dataset, using the predict data then calculates an MAPE error.

    Parameters
    ----------
    X_train: pd.DataFrame, default=X_train
        A pandas dataframe for training the model not containing the target feature.

    y_train: pd.DataFrame, default=y_train
        A pandas dataframe for training the model containing the target feature.

    X_test: pd.DataFrame, default=X_test
        A pandas dataframe for testing the model not containing the target feature.

    y_test: pd.DataFrame, default=y_test
        A pandas dataframe for testing the model containing the target feature.

    fh: int, default=24
        A period that indicates the forecast horizon while making prediction in Hours.

    summarize_period: List[int], default=[24, 48, 72]
        The period at which the window summarizer transformer will be applied and
        calculate the lag, mean and std.
        For eg. [24, 48, 72] indicate: lag of 72 period, mean and std of first
        24 period, then 48 period and last 72 period.
        Note: Period can be sort of duration, in the above example it is hours.

    model_params: Optional[Dict[str, Any]], default=None
        A dict containing the hyperparameters of the LightGBM model.
        If None is provided then model is build with default parameters.

    Returns
    -------
    float
        MAPE error is returned, this indicates the error deviation between the
        actual and predicted values. Error value needs to be between 0 and 1
        and the value should be lower for better results.

    """

    # Building the model forecasting pipeline
    logger.info(
        f"Building LightGBM model with following parameters: {model_params} "
        f"and summarize period: {summarize_period}"
    )

    pipe = build_lightgbm_model(
        summarize_period=summarize_period,
        model_params=model_params,
    )

    # Training and prediction using the forecasting pipeline and
    # getting the error deviation
    fh = np.arange(fh) + 1

    pipe_return = pipe.fit(y=y_train, X=X_train, fh=fh)
    y_pred = pipe_return.predict(X=X_test)

    error = mean_absolute_percentage_error(
        y_true=y_test, y_pred=y_pred, symmetric=False
    )

    logger.info(f"Model is been evaluated and the MAPE error is: {error}")

    return error


def objective(trial: optuna.trial.Trial):

    lgbm_params = {
        "learning_rate": trial.suggest_float("learning_rate", 1e-3, 1e-1, log=True),
        "num_leaves": trial.suggest_int("num_leaves", 2, 256),
        "min_data_in_leaf": trial.suggest_int("min_data_in_leaf", 1, 100),
        "bagging_fraction": trial.suggest_float("bagging_fraction", 0.1, 1.0),
        "colsample_bytree": trial.suggest_float("colsample_bytree", 0.1, 1.0),
        "n_estimators": trial.suggest_int("n_estimators", 500, 500),
        "bagging_freq": trial.suggest_int("bagging_freq", 1, 1),
    }

    error = model_tuning(
        X_train=X_train,
        y_train=y_train,
        X_test=X_test,
        y_test=y_test,
        fh=args.fh,
        summarize_period=args.summarize_period,
        model_params=lgbm_params,
    )

    return error


@log_exception(logger=logger)
@validate_call
def hyperparameter_tuning(filepath: Path = args.filepath):
    """
    This function tunes the model for finding the best hyperparameters and saves the
    final best configuration locally as a JSON file and in WandB artifact.

    Parameters
    ----------
    filepath: Path, default='data/assets/hyperparameter/best_config.json'
        The filepath to save the configuration as JSON file, the path needs to have a
        .json extension at the end of the filename.
    """

    with init_wandb_run(
        run_name="get_best_hyperparameter",
        job_type="model_tuning",
        group="model",
    ) as run:

        logger.info(
            f"Details of the wandb run: run ID: {run.id}, run name: {run.name}, "
            f"project name: {run.project}, entity: {run.entity}, run group: "
            f"{run.group}, run save directory: {run.dir}, run URL: {run.url}"
        )

        # Using optuna to find the best hyperparameters and tracking it with WandB
        study = optuna.create_study(direction="minimize")
        study.optimize(objective, n_trials=args.n_trials)

        # saving the best params locally and also logging it as an artifact
        Path(filepath).parent.mkdir(parents=True, exist_ok=True)
        save_json_data(data=study.best_params, filepath=filepath)

        artifact = wandb.Artifact(
            name="best_config", type="model", metadata=study.best_params
        )
        artifact.add_file(local_path=filepath)
        run.log_artifact(artifact)

        logger.info(f"Best params for the LightGBM model is: {study.best_params}")
        logger.info(
            f"Hyperparameters has been saved to filepath: {filepath} and "
            "Artifact best_config has been logged successfully"
        )

        run.summary["best_mape"] = study.best_value

        run.log(
            {
                "optimization_history_plot": plot_optimization_history(study),
                "param_importance_plot": plot_param_importances(study),
                "features_slice_plot": plot_slice(study),
            }
        )

        run.finish()


if __name__ == "__main__":

    hyperparameter_tuning()



import argparse
from pathlib import Path

import numpy as np
import pandas as pd
from pydantic import validate_call
from sktime.performance_metrics.forecasting import (
    mean_absolute_percentage_error,
    mean_squared_percentage_error,
)

from energy_consumption_forecasting.exceptions import log_exception
from energy_consumption_forecasting.inference_pipeline.batch_data import (
    get_batch_data_from_hopsworks,
)
from energy_consumption_forecasting.inference_pipeline.utils import (
    get_gcs_bucket,
    read_blob_from_bucket,
    write_blob_to_bucket,
)
from energy_consumption_forecasting.logger import get_logger
from energy_consumption_forecasting.utils import get_env_var

logger = get_logger(name=Path(__file__).name)

import hopsworks


@log_exception(logger=logger)
@validate_call
def compute_performance_metrics(
    feature_view_ver: int = 1,
    feature_view_name: str = "denmark_energy_consumption_view",
    target_feature: str = "consumption_kwh",
) -> None:
    """
    
    """

    # Getting the bucket object from GCS and reading the cached prediction data
    logger.info("Connecting to the GCS bucket and reading the cached prediction data.")
    bucket = get_gcs_bucket()
    logger.info("Connection to the GCS bucket is established.")
    cached_prediction = read_blob_from_bucket(
        bucket=bucket,
        blob_name="cached_prediction.parquet",
    )

    if cached_prediction is None or len(cached_prediction) == 0:
        logger.info(
            "Cached prediction does not exist or data is missing, "
            "exiting the process."
        )
        return None

    logger.info("Successfully loaded cached prediction data.")

    # Connecting to hopsworks feature store and getting data with
    # similar date range of the prediction
    project = hopsworks.login(
        project=get_env_var(key="FEATURE_STORE_PROJECT_NAME"),
        api_key_value=get_env_var(key="FEATURE_STORE_API_KEY"),
    )
    feature_store = project.get_feature_store()
    logger.info(
        f'Connected to Hopsworks: Project Name "{project.name}" and '
        f'Project URL: "{project.get_url()}"'
    )

    pred_start_datetime = cached_prediction.index.get_level_values(
        level="datetime_dk"
    ).min()
    pred_end_datetime = cached_prediction.index.get_level_values(
        level="datetime_dk"
    ).max()
    logger.info(
        "Getting data from hopsworks feature store between "
        f'"{pred_start_datetime}" and "{pred_end_datetime}".'
    )
    _, ground_truth = get_batch_data_from_hopsworks(
        feature_store=feature_store,
        start_datetime=pred_start_datetime.to_timestamp(),
        # Adding 1 hour for less then condition in end_datetime
        end_datetime=(pred_end_datetime + 1).to_timestamp(),
        feature_view_ver=feature_view_ver,
        feature_view_name=feature_view_name,
        target_feature=target_feature,
    )

    if len(ground_truth) == 0:
        logger.info(
            "Ground truth data does not exist or data is missing, "
            "exiting the process."
        )
        return None

    gt_start_datetime = (
        ground_truth.index.get_level_values(level="datetime_dk").min().to_timestamp()
    )
    gt_end_datetime = (
        ground_truth.index.get_level_values(level="datetime_dk").max().to_timestamp()
    )
    logger.info(
        "Ground truth data received from hopsworks feature store, data is between "
        f'"{gt_start_datetime}" and "{gt_end_datetime}".'
    )

    logger.info("Successfully loaded ground truth data.")

    # Merging both the dataframe for computing
    logger.info("Computing performance metrics.")
    cached_prediction = cached_prediction.rename(
        columns={"consumption_kwh": "consumption_kwh_pred"}
    )
    ground_truth = ground_truth.rename(
        columns={"consumption_kwh": "consumption_kwh_gt"}
    )
    cached_prediction["consumption_kwh_gt"] = np.nan
    cached_prediction.update(ground_truth)

    # Computing MAPE and RMSPE performance metric between prediction and ground truth
    cached_prediction = cached_prediction.dropna(subset=["consumption_kwh_gt"])
    if len(cached_prediction) == 0:
        logger.info(
            "Ground truth data does not exist or data is missing, "
            "exiting the process."
        )
        return None

    mape = (
        cached_prediction.groupby("datetime_dk")
        .apply(
            lambda data: mean_absolute_percentage_error(
                y_true=data["consumption_kwh_gt"],
                y_pred=data["consumption_kwh_pred"],
                symmetric=False,
            )
        )
        .rename("mape")
        .to_frame()
    )

    rmspe = (
        cached_prediction.groupby("datetime_dk")
        .apply(
            lambda data: mean_squared_percentage_error(
                y_true=data["consumption_kwh_gt"],
                y_pred=data["consumption_kwh_pred"],
                square_root=True,
            )
        )
        .rename("rmspe")
        .to_frame()
    )

    metric_result = pd.merge(left=mape, right=rmspe, on="datetime_dk")
    logger.info("Performance metrics of MAPE and RMSPE successfully computed.")

    # Saving the metrics dataframe in the GCS bucket as a parquet file
    logger.info("Saving performance metrics dataframe in GCS bucket.")
    write_blob_to_bucket(
        bucket=bucket, blob_name="performance_metrics.parquet", data=metric_result
    )
    logger.info(
        'Performance metrics file "performance_metrics.parquet" saved in GCS bucket.'
    )

    # Saving the ground truth dataframe in the GCS bucket as a parquet file
    logger.info("Saving ground truth dataframe in GCS bucket.")
    ground_truth = ground_truth.rename(
        columns={"consumption_kwh_gt": "consumption_kwh"}
    )
    write_blob_to_bucket(
        bucket=bucket, blob_name="ground_truth.parquet", data=ground_truth
    )
    logger.info('Ground truth file "ground_truth.parquet" saved in GCS bucket.')


if __name__ == "__main__":

    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--views_name",
        type=str,
        default="denmark_energy_consumption_view",
        help="Name of feature view within the feature group, needs to be in string format.",
    )

    parser.add_argument(
        "--views_ver",
        type=int,
        default=1,
        help="Version of feature view within the feature group, "
        "needs to be in integer format.",
    )

    parser.add_argument(
        "--target_feature",
        type=str,
        default="consumption_kwh",
        help="Name of target feature, needs to be in string format.",
    )

    args = parser.parse_args()

    compute_performance_metrics(
        feature_view_ver=args.views_ver,
        feature_view_name=args.views_name,
        target_feature=args.target_feature,
    )



import datetime
import os
from json import dump
from pathlib import Path
from typing import Any, Dict, Optional, Tuple

import httpx
import pandas as pd
from pydantic import HttpUrl, validate_call

from energy_consumption_forecasting.exceptions import log_exception
from energy_consumption_forecasting.logger import get_logger
from energy_consumption_forecasting.utils import get_env_var, save_json_data

logger = get_logger(name=Path(__file__).name)
ROOT_DIRPATH = Path(get_env_var(key="PROJECT_ROOT_DIR_PATH", default_value="."))


@log_exception(logger=logger)
@validate_call
def get_extraction_datetime(
    start_date_time: datetime.datetime,
    end_date_time: datetime.datetime,
) -> Tuple[str, str]:
    """
    This Function formats the datetime in a format that the API will accept.

    Parameters
    ----------
    start_date_time: datetime.datetime
        A starting date and time for extracting the data in datatype of
        datetime.datetime.

    end_date_time: datetime.datetime
        A ending date and time for extracting the data in datatype of
        datetime.datetime.

    Returns
    -------
    start_date_time: str
        A string of formatted starting date and time.

    end_date_time: str
        A string of formatted ending date and time.
    """

    # Checking if end date is greater than start date
    if start_date_time > end_date_time:
        raise Exception("End date needs to be greater than the start date")

    # Converting the date format for API query and
    # increasing the end date by 1 day as per the API guide (off-by-one error)
    start_date_time = start_date_time.strftime("%Y-%m-%dT%H:%M")
    end_date_time = (end_date_time + datetime.timedelta(days=1)).strftime(
        "%Y-%m-%dT%H:%M"
    )

    return start_date_time, end_date_time


@log_exception(logger=logger)
@validate_call
def extract_dataset_from_api(
    start_date_time: datetime.datetime,
    end_date_time: datetime.datetime,
    sort_data_asc: bool = True,
    dataset_name: str = "ConsumptionIndustry",
    base_url: HttpUrl = "https://api.energidataservice.dk/dataset/",
    meta_url: HttpUrl = "https://api.energidataservice.dk/meta/dataset/",
    save_dataset_metadata: bool = True,
) -> Optional[Tuple[pd.DataFrame, Dict[str, Any] | Path, Path]]:
    """
    This function extracts data using the API from the
    Denmark Energy Data Service website: "https://www.energidataservice.dk/".

    Parameters
    ----------
    start_date_time: datetime.datetime
        A starting date and time for extracting the data in datatype of
        datetime.datetime.

    end_date_time: datetime.datetime
        A ending date and time for extracting the data in datatype of
        datetime.datetime.

    sort_data_asc: bool, default=True
        Sort the data using the UTC datetime column, by default data is sorted
        in ascending order.

    dataset_name: str, default="ConsumptionIndustry"
        A string containing the dataset name that needs to be extracted from
        the website.
        You can find the dataset name under the section of Additional Info in
        Alias tag.

    base_url: str, default="https://api.energidataservice.dk/dataset/"
        The base URL for the data.

    meta_url: str, default="https://api.energidataservice.dk/meta/dataset/"
        The base URL for the metadata.

    save_dataset_metadata: bool, default=True
        Whether to save the dataset as a CSV file and metadata as a JSON file.

    Returns
    -------
    pd.DataFrame, Dict[str, Any] or Path, Path
        A tuple containing the dataset in pandas DataFrame and a Dict containing
        the metadata of the dataset.
        If save_dataset_metadata parameter is True, then filepath for both the
        data is returned as string along with the DataFrame and Dict.
    """

    data_url = f"{base_url}{dataset_name}?"
    meta_url = f"{meta_url}{dataset_name}?"
    sort = "HourUTC" if sort_data_asc else "HourUTC%20DESC"

    # Formatting the dates for the API parameters
    start, end = get_extraction_datetime(
        start_date_time=start_date_time, end_date_time=end_date_time
    )

    # Creating the parameters for the API request
    params = {
        "offset": 0,
        "start": start,
        "end": end,
        "sort": sort,
    }

    # Calling the API requests for dataset and metadata
    with httpx.Client() as client:
        logger.info(
            f"Sending API get request to: {data_url} and {meta_url} "
            f"with parameters: {params}."
        )

        data_response = client.get(url=data_url, params=params, timeout=None)
        meta_response = client.get(url=meta_url, timeout=None)

        logger.info(
            "Connection to the dataset API is done and response "
            f"received with status code: {data_response.status_code}."
        )
        logger.info(
            "Connection to the metadata API is done and response "
            f"received with status code: {meta_response.status_code}."
        )

        # Getting the data and metadata in json format for further processing
        json_data = data_response.json()
        json_meta = meta_response.json()

    # Getting the dataset from the JSON data and converting into dataframe
    json_data = json_data.get("records")
    dataset_df = pd.DataFrame.from_records(json_data)

    if save_dataset_metadata:
        data_dir = ROOT_DIRPATH / "data" / "raw_data"
        if not os.path.isdir(data_dir):
            os.makedirs(data_dir)

        start = start.replace(":", "-")
        end = end.replace(":", "-")
        data_filepath = data_dir / f"{dataset_name}_{start}_{end}.csv"
        meta_filepath = data_dir / f"{dataset_name}_metadata.json"

        logger.info(
            f'Saving the dataset "{data_filepath.name}" in '
            f'directory: "{data_dir.absolute()}".'
        )
        logger.info(
            f'Saving the metadata "{meta_filepath.name}" in '
            f'directory: "{data_dir.absolute()}".'
        )

        # Saving the dataset as a csv file and
        # meta data as JSON file in data directory
        dataset_df.to_csv(path_or_buf=data_filepath, index=False)
        save_json_data(data=json_meta, filepath=meta_filepath)

        logger.info(f'Dataset has been saved in csv file "{data_filepath.name}".')
        logger.info(f'Metadata has been saved in json file "{meta_filepath.name}".')

        return dataset_df, json_meta, data_filepath, meta_filepath

    return dataset_df, json_meta
